# coding: utf-8
#/*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/


__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "09/08/2018"


import numpy
import logging
import typing
import json
import io
import os
import h5py
from glob import glob
from silx.io.url import DataUrl
from tomwer.core.utils.locker import FileLockerManager, FileLockerContext
from tomwer.core.utils.ftseriesutils import orderFileByLastLastModification

logger = logging.getLogger(__name__)


class TomwerScanBase:
    """
    Simple interface to extend the tomoscan.TomwerScanBase with
    specific functions
    """

    _DICT_TOMO_RP_KEY = "ftseries_recons_params"

    _DICT_LAMINO_RP_KEY = "lamino_recons_params"

    _DICT_AXIS_KEYS = "axis_recons_params"

    _DICT_PROCESS_INDEX_KEY = "next_process_index"

    VALID_RECONS_EXTENSION = '.edf', '.npy', '.npz', '.hdf5'

    def __init__(self, overwrite_proc_file=False):
        self._reconstructions = []

        self._ftseries_params = None
        """ft parameters"""
        # TODO: this information can be used on ftseries to know if this is
        # the first time we are running a reconstruction or not
        self._lamino_params = None
        """Set of reconstructions parameters for laminography"""
        self._axis_params = None
        """Axis parameters"""
        self._process_file = None
        """file storing processes applied on the scan, with their configuration
        and result"""
        self._cache_proj_urls = None
        """cache for the projection urls"""
        self._cache_radio_axis = {}
        """cache for the radio axis. Key is tuple (mode, nearest), value is
        (url1, url2)"""
        self._normed_darks = None
        """darks normed. Dict: key is the index in the sequence and value is
        the mean or median dark"""
        self._normed_flats = None
        """flats normed. Dict: key is the index in the sequence and value is
        the mean or median dark"""
        self._notify_ffc_rsc_missing = True
        """Should we notify the user if ffc fails because cannot find dark or
        flat. Used to avoid several warnings. Only display one"""

    def _init_index_process_file(self, overwrite_proc_file=False):
        if not overwrite_proc_file and self.process_file is not None and os.path.exists(self.process_file):
            with h5py.File(self.process_file, mode='r', swmr=True) as h5s:
                self._process_index = len(h5s.items())
        else:
            self._process_index = 0

    def clear_caches(self):
        self._cache_proj_urls = None
        self._notify_ffc_rsc_missing = True

    @property
    def normed_darks(self):
        return self._normed_darks

    def set_normed_darks(self, darks):
        self._normed_darks = darks

    @property
    def normed_flats(self):
        return self._normed_flats

    def set_normed_flats(self, flats):
        self._normed_flats = flats

    def _flat_field_correction(self, data, index: typing.Union[int, None],
                               dark, flat1, flat2, index_flat1: int,
                               index_flat2: int):
        assert type(data) is numpy.ndarray
        can_process = True

        if dark is None:
            if self._notify_ffc_rsc_missing:
                logger.error(
                    'cannot make flat field correction, dark not found')
            can_process = False

        if dark is not None and dark.ndim != 2:
            logger.error(
                'cannot make flat field correction, dark should be of '
                'dimension 2')
            can_process = False

        if flat1 is None:
            if self._notify_ffc_rsc_missing:
                logger.error(
                    'cannot make flat field correction, flat not found')
            can_process = False
        else:
            if flat1.ndim != 2:
                logger.error(
                    'cannot make flat field correction, flat should be of '
                    'dimension 2')
                can_process = False
            if flat2 is not None and flat1.shape != flat2.shape:
                logger.error(
                    'the tow flats provided have different shapes.')
                can_process = False

        if dark is not None and flat1 is not None and dark.shape != flat1.shape:
            logger.error('Given dark and flat have incoherent dimension')
            can_process = False

        if dark is not None and data.shape != dark.shape:
            logger.error('Image has invalid shape. Cannot apply flat field'
                          'correction it')
            can_process = False

        if can_process is False:
            self._notify_ffc_rsc_missing = False
            return data

        if flat2 is None:
            flat_value = flat1
        else:
            # compute weight and clip it if necessary
            if index is None:
                w = 0.5
            else:
                w = (index - index_flat1) / (index_flat2 - index_flat1)
                w = min(1, w)
                w = max(0, w)
            flat_value = flat1 * w + flat2 * (1-w)

        div = flat_value - dark
        div[div == 0] = 1
        return (data - dark) / div

    def acquire_process_file_lock(self):
        """create a FileLocker context manager to insure safe write to the
        process file"""
        if self.process_file is None:
            raise ValueError('No processing file defined')
        file_locker = FileLockerManager().get_locker(file_=self.process_file)
        return FileLockerContext(file_locker)

    @property
    def reconstructions(self):
        """list of reconstruction files"""
        return self._reconstructions

    @reconstructions.setter
    def reconstructions(self, reconstructions):
        self._reconstructions = reconstructions

    @property
    def ftseries_recons_params(self):
        return self._ftseries_params

    @ftseries_recons_params.setter
    def ftseries_recons_params(self, recons_params):
        self._ftseries_params = recons_params

    @property
    def lamino_recons_params(self):
        return self._lamino_params

    @lamino_recons_params.setter
    def lamino_recons_params(self, recons_params):
        self._lamino_params = recons_params

    @property
    def axis_params(self):
        return self._axis_params

    @axis_params.setter
    def axis_params(self, parameters):
        self._axis_params = parameters

    # TODO: change name. Should be generalized to return Dataurl
    def getReconstructedFilesFromParFile(self, with_index):
        raise NotImplementedError('Base class')

    def pop_process_index(self) -> int:
        """Return and lock the next free process index"""
        process_index = self._process_index
        self._process_index += 1
        return process_index

    def getRadiosForAxisCalc(self, mode, nearest: bool = True,
                             use_cache: bool = True) -> tuple:
        """
        Return the radios for axis calculation and the requested mode.

        :param: angles we want to use for COR calculation. Can be 0-180,
                 90-180 or manual. If manual will return the 'most' appropriate
        :type: CorAngleMode
        :param nearest: if True then, pick the closest angle from the requested
                        one. If False, return (None, None) if the the angles
                        does not exists.
        :type: bool
        :return: couple of `opposite` radios that can be used for axis
                 calculation.
        :rtype: tuple(AxisResource, AxisResource)
        """
        if not use_cache:
            del self._cache_radio_axis[(mode, nearest)]

        def compute():
            from ..process.reconstruction.axis.params import AxisResource  # avoid cyclic import
            from ..process.reconstruction.axis.anglemode import CorAngleMode

            _mode = CorAngleMode.from_value(mode)
            if self.path is None:
                return None, None

            if hasattr(self, 'projections_with_angle'):
                radios_with_angle = self.projections_with_angle()
            else:
                radios_with_angle = self.get_proj_angle_url()
            if _mode is CorAngleMode.use_0_180:
                couples = ((0, 180),)
            elif _mode is CorAngleMode.use_90_270:
                couples = ((90, 270),)
            else:
                couples = ((0, 180), (90, 270))

            for couple in couples:
                if(couple[0] in radios_with_angle.keys()
                        and couple[1] in radios_with_angle.keys()):
                    radio_0 = AxisResource(radios_with_angle[couple[0]])
                    radio_1 = AxisResource(radios_with_angle[couple[1]])
                    return radio_0, radio_1

            def find_nearest(angles, angle):
                if len(angles) == 0:
                    return None
                angles = numpy.asarray(angles)
                dist = numpy.abs(angles - angle)
                idx = dist.argmin()
                if isinstance(idx, numpy.ndarray):
                    idx = idx[0]
                return angles[idx]

            if nearest is True:
                angles = []
                # filter str value (0(1)...)
                [angles.append(value) for value in radios_with_angle.keys()
                 if numpy.issubdtype(type(value), numpy.number)]
                nearest_c1 = find_nearest(angles=angles, angle=couples[0][0])
                nearest_c2 = find_nearest(angles=angles, angle=couples[0][1])
                if nearest_c1 is not None and nearest_c2 is not None:
                    radio_0 = AxisResource(radios_with_angle[nearest_c1])
                    radio_1 = AxisResource(radios_with_angle[nearest_c2])
                    return radio_0, radio_1
            return None, None

        if (mode, nearest) not in self._cache_radio_axis:
            radio_0, radio_1 = compute()
            self._cache_radio_axis[(mode, nearest)] = (radio_0, radio_1)
        return self._cache_radio_axis[(mode, nearest)]

    def flat_field_correction(self, data, index=None):
        """Apply flat field correction on the given data

        :param numpy.ndarray data: the data to apply correction on
        :param Uion[int, None] index: index of the data in the acquisition
                                      sequence
        :return: corrected data
        :rtype: numpy.ndarray
        """
        raise NotImplementedError('Base class')

    def getReconsParamList(self):
        """

        :return: reconstruction parameters
        :rtype: ReconsParamList
        """
        raise NotImplementedError('Base class')

    @property
    def process_file(self) -> str:
        """

        :return: file used to store the processes launch by tomwer
        """
        return self._process_file

    def to_dict(self):
        res = {}
        # ftseries reconstruction parameters
        if self.ftseries_recons_params:
            res[self._DICT_TOMO_RP_KEY] = self.ftseries_recons_params.to_dict()
        else:
            res[self._DICT_TOMO_RP_KEY] = None
        # axis reconstruction parameters
        if self.axis_params is None:
            res[self._DICT_AXIS_KEYS] = None
        else:
            res[self._DICT_AXIS_KEYS] = self.axis_params.to_dict()
        # process index
        res[self._DICT_PROCESS_INDEX_KEY] = self._process_index

        # lamino reconstruction parameters
        res[self._DICT_LAMINO_RP_KEY] = self.lamino_recons_params
        return res

    def load_from_dict(self, desc):
        from tomwer.core.process.reconstruction.ftseries.params import ReconsParams  # avoid cyclic import
        from tomwer.core.process.reconstruction.axis.params import AxisRP  # avoid cyclic import

        if isinstance(desc, io.TextIOWrapper):
            data = json.load(desc)
        else:
            data = desc
        if not (self.DICT_PATH_KEY in data and data[self.DICT_TYPE_KEY] == self._TYPE):
            raise ValueError('Description is not an EDFScan json description')

        assert self.DICT_PATH_KEY in data
        assert self._DICT_LAMINO_RP_KEY in data
        self.path = data[self.DICT_PATH_KEY]
        recons_param_data = data[self._DICT_TOMO_RP_KEY]
        if recons_param_data is not None:
            self.ftseries_recons_params = ReconsParams.from_dict(recons_param_data)
        self.lamino_recons_params = data[self._DICT_LAMINO_RP_KEY]
        axis_params = data[self._DICT_AXIS_KEYS]
        if axis_params is not None:
            self.axis_params = AxisRP.from_dict(axis_params)
        self._process_index = data[self._DICT_PROCESS_INDEX_KEY]

    def equal(self, other):
        """

        :param :class:`.ScanBase` other: instance to compare with 
        :return: True if instance are equivalent
        :note: we cannot use the __eq__ function because this object need to be
               pickable
        """
        return (
            isinstance(other, self.__class__) or isinstance(self, other.__class__) and
            self.type == other.type and
            self.ftseries_recons_params == other.ftseries_recons_params and
            self.lamino_recons_params == other.lamino_recons_params and
            self.path == other.path
        )

    def get_sinogram(self, line, progress=None):
        """
        extract the sinogram from projections

        :param line: which sinogram we want
        :type: int
        :return: sinogram from the radio lines
        :rtype: numpy.array
        """
        raise NotImplementedError('Base class')

    def __str__(self):
        raise NotImplementedError('Base class')

    @staticmethod
    def get_pyhst_recons_file(scanID):
        """Return the .par file used for the current reconstruction if any.
        Otherwise return None """
        if scanID == "":
            return None

        if scanID is None:
            raise RuntimeError('No current acquisition to validate')
        assert(type(scanID) is str)
        assert(os.path.isdir(scanID))
        folderID = os.path.basename(scanID)
        # look for fasttomo files ending by slice.par
        parFiles = glob(os.path.join(scanID + folderID) + '*_slice.par')
        if len(parFiles) > 0:
            return orderFileByLastLastModification(scanID, parFiles)[-1]
        else:
            return None

    def _deduce_transfert_scan(self, output_dir):
        """
        Create the scan that will be generated if this scan is
        copy to the output_dir

        :param str output_dir:
        """
        raise NotImplementedError('Base class')

    def get_proj_angle_url(self, use_cache: bool = True):
        """
        retrieve the url for each projections (including the alignement /
        return one) and associate to each (if possible) the angle.
        Alignment angle are set as angle (1) to specify that this is an
        alignment one.
        :param bool use_cache:
        :return: dictionary with angle (str or int) as key and url as value
        :rtype: dict
        """
        raise NotImplementedError('Base class')

    def get_reconstructions_urls(self):
        """

        :return: list of urls that contains reconstruction from nabu or pyhst
        :rtype: list
        """
        def containsDigits(input):
            return any(char.isdigit() for char in input)

        if (self.path is None) or (not os.path.isdir(self.path)):
            return []
        results = []
        # parse .npy, .npz and .edf, .vol, *.hdf5 files
        folderBasename = os.path.basename(self.path)
        pyhst_files = TomwerScanBase.get_pyhst_recons_file(self.path)
        if pyhst_files is not None:
            results.extend(
                TomwerScanBase.getReconstructedFilesFromParFile(pyhst_files,
                                                                with_index=False))
        if os.path.isdir(self.path):
            for f in os.listdir(self.path):
                if f.endswith(TomwerScanBase.VALID_RECONS_EXTENSION) and f.startswith(folderBasename) and 'slice_' in f:
                    local_str = f
                    for extension in TomwerScanBase.VALID_RECONS_EXTENSION:
                        if local_str.endswith(extension):
                            local_str = local_str.rsplit(extension, 1)[0]
                    if 'slice_' in local_str:
                        if 'slice_pag_' in local_str:
                            indexStr = local_str.split('slice_pag_')[-1].split('_')[0]
                        else:
                            indexStr = local_str.split('slice_')[-1].split('_')[0]
                        if containsDigits(indexStr):
                            gfile = os.path.join(self.path, f)
                            assert(os.path.isfile(gfile))
                            if gfile.lower().endswith('.edf'):
                                url = DataUrl(file_path=gfile, scheme='fabio')
                            elif gfile.lower().endswith(('.npy', 'npz')):
                                url = DataUrl(file_path=gfile, scheme='numpy')
                            elif gfile.lower().endswith(('.h5', '.hdf5', '.nx')):
                                entry = getattr(self, 'entry', 'entry')
                                with h5py.File(gfile, 'r', swmr=True) as h5f:
                                    if entry in h5f:
                                        url = DataUrl(file_path=gfile,
                                                      scheme='silx',
                                                      data_path='/'.join((entry, 'reconstruction', 'results', 'data'))
                                                      )
                                    else:
                                        continue
                            results.append(url)
                if f.endswith(".vol"):
                    url = DataUrl(file_path=os.path.join(self.path, f),
                                  scheme='tomwer')
                    results.append(url)
        return results


class _TomwerBaseDock(object):
    """
    Internal class to make difference between a simple TomoBase output and
    an output for a different processing (like scanvalidator.UpdateReconsParam)
    """
    def __init__(self, tomo_instance):
        self.__instance = tomo_instance

    @property
    def instance(self):
        return self.__instance



# coding: utf-8
#/*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/
"""module defining functions and class to communicate a status to the
orange-server.
The orange-server is used to display the advancement of the workflow.
"""

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "28/04/2017"


try:
    import graypy
    has_graypy = True
except:
    has_graypy = False
from .config import grayport_port, grayport_host
import os
import socket
import logging

_logger = logging.getLogger(__name__)

global _INFO_UNKNOWN_HOST_NAME
_INFO_UNKNOWN_HOST_NAME = False
"""Cache variable to avoid several log about this information"""


class OWClient(object):
    """Orange widget Client can emit information about his advancement
    """

    WORKFLOW_INFO = 'workflow'  # general information about a workflow
    SCAN_INFO = 'scan'  # information focus on a scan

    _log_graylog_err = True
    """insure log on GELF handler is only displayed one"""

    def __init__(self, loggers):
        assert loggers is not None
        if OWClient._knows_hostname(hostname=grayport_host) is False:
            global _INFO_UNKNOWN_HOST_NAME
            if _INFO_UNKNOWN_HOST_NAME is False:
                _logger.warning('unknow host %s' % grayport_host)
                _INFO_UNKNOWN_HOST_NAME = True
            return
        if type(loggers) not in (list, tuple):
            loggers = (loggers, )
        for logger in loggers:
            if os.environ.get('ORANGE_WEB_LOG', 'True') == 'True':
                try:
                    self.graylogHandler = graypy.GELFHandler(grayport_host, grayport_port)
                except:
                    if OWClient._log_graylog_err:
                        logger.warning('Fail to create GELFHandler. Won\'t report log message')
                        OWClient._log_graylog_err = False
                else:
                    logger.addHandler(self.graylogHandler)
                    logger.debug('- add graypy handler')
            else:
                info = 'No log will be send to graylog.'
                info += 'ORANGE_WEB_LOG variable is setted to False'
                logger.debug(info)

    @staticmethod
    def _knows_hostname(hostname):
        if hostname is None:
            return False
        try:
            socket.gethostbyname(hostname)
            return True
        except socket.error:
            return False



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
contains gui relative to axis calculation using radios
"""

__authors__ = ["C. Nemoz", "H. Payno"]
__license__ = "MIT"
__date__ = "25/02/2019"

from silx.gui import qt
from tomwer.core.log import TomwerLogger
from tomwer.core.utils import image
import numpy
import enum
import os
from .CompareImages import MODE_RAW_COMPARISON
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.synctools.axis import QAxisRP
from tomwer.core.process.reconstruction.axis.mode import AxisMode
from tomwer.core.process.reconstruction.axis.anglemode import CorAngleMode
from tomwer.gui.UrlSelectionTable import ColumnMode, UrlSelectionDialog
from tomwer.core.process.reconstruction.axis.params import AxisCalculationInput
from tomwer.core.process.reconstruction.axis.projectiontype import ProjectionType
from collections import OrderedDict
import scipy.signal
from .CompareImages import CompareImages
from tomwer.gui.utils.buttons import PadlockButton

_logger = TomwerLogger(__name__)


class RadioAxisWindow(qt.QMainWindow):
    """
    QMainWindow for defining the rotation axis

    .. snapshotqt:: img/AxisWindow.png

        from tomwer.gui.ftserie.axis import AxisWindow
        from tomwer.synctools.ftseries import QReconsParams
        import scipy.misc
        import scipy.ndimage

        recons_params = QReconsParams()
        widget = AxisWindow(recons_params)
        imgA = scipy.misc.ascent()
        imgB = scipy.ndimage.affine_transform(imgA, (1, 1), offset=(0, 10))

        widget.setImages(imgA=imgA, imgB=imgB, flipB=False)
        widget.show()

    :raises ValueError: given axis is not an instance of _QAxisRP
    """

    sigAxisEditionLocked = qt.Signal(bool)
    """Signal emitted when the status of the reconstruction parameters edition
    change"""

    sigLockModeChanged = qt.Signal(bool)
    """signal emitted when the lock on the mode change"""

    sigPositionChanged = qt.Signal(tuple)
    """signal emitted when the center of rotation center change"""

    def __init__(self, axis, parent=None):
        qt.QMainWindow.__init__(self, parent)
        if isinstance(axis, QAxisRP):
            self.__recons_params = axis
        else:
            raise TypeError('axis should be an instance of _QAxisRP')

        self._imgA = None
        self._imgB = None
        self._shiftedImgA = None
        self._flipB = True
        """Option if we want to flip the image B"""
        self.__callback = None
        self._scan = None
        self._axis_params = None
        self._lastManualFlip = None
        """Cache for the last user entry for manual flip"""

        self.setWindowFlags(qt.Qt.Widget)
        self._plot = CompareImages(parent=self)
        _mode = MODE_RAW_COMPARISON
        self._plot.setVisualizationMode(_mode)
        self.setCentralWidget(self._plot)

        self._dockWidgetCtrl = qt.QDockWidget(parent=self)
        self._dockWidgetCtrl.layout().setContentsMargins(0, 0, 0, 0)
        self._dockWidgetCtrl.setFeatures(qt.QDockWidget.DockWidgetMovable)
        self._controlWidget = _AxisManual(parent=self, reconsParams=self.__recons_params)
        self._dockWidgetCtrl.setWidget(self._controlWidget)
        self.addDockWidget(qt.Qt.RightDockWidgetArea, self._dockWidgetCtrl)

        # expose API
        self.getXShift = self._controlWidget.getXShift
        self.getYShift = self._controlWidget.getYShift
        self._setShift = self._controlWidget.setShift
        self.setXShift = self._controlWidget.setXShift
        self.setYShift = self._controlWidget.setYShift
        self.getShiftStep = self._controlWidget.getShiftStep
        self.setShiftStep = self._controlWidget.setShiftStep
        self.getAxis = self._controlWidget.getAxis
        self.getMode = self._controlWidget.getMode

        # signal / slot connection
        self._controlWidget.sigShiftChanged.connect(self._updateShift)
        self._controlWidget.sigShiftChanged.connect(self._corChanged)
        self._controlWidget.sigRoiChanged.connect(self._updateShift)
        self._controlWidget.sigAuto.connect(self._updateAuto)
        self._controlWidget.sigModeChanged.connect(self.setMode)
        self.__recons_params.sigChanged.connect(self._reset_scan)
        self._plot.flipToggled.connect(self._flipChanged)

        # adapt gui to the axis value
        self.setReconsParams(axis=self.__recons_params)

        # set  up
        self._plot.setRadio2FlipVisible(self.getMode() == AxisMode.manual)

    def _corChanged(self):
        self.sigPositionChanged.emit((self.getXShift(), self.getYShift()))

    def getPlot(self):
        return self._plot

    def getSettingsWidget(self):
        return self._controlWidget

    def setMode(self, mode):
        """
        Define the mode to use for radio axis

        :param mode:
        :return:
        """
        mode = AxisMode.from_value(mode)
        mode_is_manual = mode == AxisMode.manual
        self._plot.setRadio2FlipVisible(mode_is_manual)
        # the algorithm are taking flip images so the display should be
        # consistent
        if mode_is_manual:
            if self._lastManualFlip is None:
                self._lastManualFlip = self._plot.isRadio2Flip()
            else:
                self._setRadio2Flip(self._lastManualFlip)
        else:
            self._setRadio2Flip(checked=True)

        old_control = self._controlWidget.blockSignals(True)
        self._controlWidget.setMode(mode)
        self._controlWidget.blockSignals(old_control)

    def _lockModeChanged(self, lock, disable_other_mode_lock=False):
        self._lockBut.setEnabled(not lock)
        self._lockLabel.setEnabled(not lock)
        if disable_other_mode_lock is True:
            self._controlWidget.setEnabled(not lock)
        self.sigLockModeChanged.emit(lock)

    def getROI(self):
        if self.getMode() == AxisMode.manual:
            return self._controlWidget.getROI()
        else:
            return None

    def _computationRequested(self):
        self.sigComputationRequested.emit()

    def setLocked(self, locked):
        old = self.blockSignals(True)
        if self._axis_params.mode not in (AxisMode.read, AxisMode.manual):
            self._axis_params.mode = AxisMode.manual
        self._controlWidget.setLocked(locked)
        self.blockSignals(old)
        self.sigAxisEditionLocked.emit(locked)

    def isModeLock(self):
        return self._controlWidget.isModeLock()

    def _validated(self):
        """callback when the validate button is activated"""
        self.sigApply.emit()

    def _setRadio2Flip(self, checked):
        self._plot.setRadio2Flip(checked)

    def _flipChanged(self, checked):
        if self.getMode() == AxisMode.manual:
            self._lastManualFlip = self._plot.isRadio2Flip()

        if checked == self._flipB:
            return
        else:
            self._flipB = checked
            self._updatePlot()

    def setReconsParams(self, axis):
        """

        :param AxisRP axis: axis to edit
        :return: 
        """
        assert isinstance(axis, QAxisRP)
        self._axis_params = axis
        old = self.blockSignals(True)
        self.resetShift()
        self._controlWidget.setAxis(axis)
        self.blockSignals(old)

    def setScan(self, scan):
        """
        Update the interface concerning the given scan. Try to display the
        radios for angle 0 and 180.

        :param scan: scan for which we want the axis updated.
        :type scan: Union[str, tomwer.core.scan.TomoBase]
        """
        self.clear()
        _scan = scan
        if type(scan) is str:
            try:
                _scan = ScanFactory.create_scan_object(scan)
            except ValueError:
                raise ValueError('Fail to discover a valid scan in %s' % scan)
        elif not isinstance(_scan, TomwerScanBase):
            raise ValueError('type of %s (%s) is invalid, scan should be a '
                             'file/dir path or an instance of ScanBase' % (
                             scan, type(scan)))
        assert isinstance(_scan, TomwerScanBase)

        if _scan.axis_params is None:
            _scan.axis_params = QAxisRP()

        if self._scan is not None:
            self._scan.axis_params.sigAxisUrlChanged.disconnect(self._updatePlot)
        self._scan = _scan
        self._scan.axis_params.sigAxisUrlChanged.connect(self._updatePlot)
        self._updatePlot()
        self._controlWidget.setScan(scan=self._scan)
        self._updatePlot()

    def _updatePlot(self):
        if self._scan is None:
            return

        coreAngleMode = CorAngleMode.from_value(self.__recons_params.angle_mode)
        axis_rp = self._scan.axis_params
        if (coreAngleMode is CorAngleMode.manual_selection and axis_rp.n_url() == 2):
            pass
        else:
            res = self._scan.getRadiosForAxisCalc(mode=coreAngleMode)
            axis_rp.axis_url_1 = res[0]
            axis_rp.axis_url_2 = res[1]

        if axis_rp.n_url() < 2:
            _logger.error('Fail to detect radio for axis calculation')
        elif axis_rp.axis_url_1.url:
            # if necessary normalize data
            axis_rp.axis_url_1.normalize_data(self._scan, log_=False)
            axis_rp.axis_url_2.normalize_data(self._scan, log_=False)

            paganin = self.__recons_params.paganin_preproc
            # check if normed
            if paganin:
                imgA = axis_rp.axis_url_1.normalized_data_paganin
                imgB = axis_rp.axis_url_2.normalized_data_paganin
            else:
                imgA = axis_rp.axis_url_1.normalized_data
                imgB = axis_rp.axis_url_2.normalized_data
            assert imgA is not None
            assert imgB is not None
            self.setImages(imgA=imgA, imgB=imgB)
        else:
            _logger.error('fail to find radios for angle 0 and 180. Unable to '
                          'update axis gui')

    def clear(self):
        if self._scan is not None:
            self._scan.axis_params.sigAxisUrlChanged.disconnect(self._updatePlot)
        self._scan = None

    def _reset_scan(self):
        if self._scan:
            self.setScan(scan=self._scan)

    def _showImgBBeFlipped(self):
        return self._plot.getCompareToolBar().isRadio2Flip()

    def setImages(self, imgA, imgB):
        """

        :warning: does not reset the shift when change images

        :param numpy.array imgA: first image to compare. Will be the one shifted
        :param numpy.array imgB: second image to compare
        :param bool flipB: True if the image B has to be flipped
        :param bool paganin: True to apply paganin phase retrieval
        """
        flipB = self._showImgBBeFlipped()
        assert imgA is not None
        assert imgB is not None
        _imgA = imgA
        _imgB = imgB

        if _imgA.shape != _imgB.shape:
            _logger.error("The two provided images have incoherent shapes "
                          "(%s vs %s)" % (_imgA.shape, _imgB.shape))
        elif _imgA.ndim != 2:
            _logger.error("Image shape are not 2 dimensional")
        else:
            self._flipB = flipB
            self._imgA = _imgA
            self._imgB = _imgB

            self._controlWidget._roiControl.setLimits(width=self._imgA.shape[1],
                                                      height=self._imgA.shape[
                                                          0])
            self._updateShift()

    def _updateShift(self, xShift=None, yShift=None):
        if self._imgA is None or self._imgB is None:
            return
        xShift = xShift or self.getXShift()
        yShift = yShift or self.getYShift()

        _imgA, _imgB = self._getRawImages()
        # apply shift
        if xShift == 0.0 and yShift == 0.0:
            self._shiftedImgA = _imgA
        else:
            try:
                self._shiftedImgA = image.shift_img(data=_imgA,
                                                    dx=self.getXShift(),
                                                    dy=self.getYShift(),
                                                    cval=_imgA.min())
            except ValueError as e:
                _logger.warning(e)
        self._plot.setData(image1=self._shiftedImgA, image2=_imgB)

    def _getRawImages(self):

        def selectROI(data, width, height):
            x_center = data.shape[1] // 2
            y_center = data.shape[0] // 2
            x_min = x_center - width // 2
            x_max = x_center + width // 2
            y_min = y_center - height // 2
            y_max = y_center + height // 2
            return data[y_min:y_max, x_min:x_max]

        # get images and apply ROI if any
        _roi = self.getROI()
        _imgA = self._imgA
        _imgB = self._imgB
        if self._flipB is True:
            _imgB = numpy.fliplr(_imgB)
        if _roi is not None:
            assert type(_roi) is tuple
            _imgA = selectROI(_imgA, width=_roi[0], height=_roi[1])
            _imgB = selectROI(_imgB, width=_roi[0], height=_roi[1])
        return _imgA, _imgB

    def _updateAuto(self):
        _imgA, _imgB = self._getRawImages()
        correlation = scipy.signal.correlate2d(in1=_imgA, in2=_imgB)
        y, x = numpy.unravel_index(numpy.argmax(correlation), correlation.shape)
        self._setShift(x=x, y=y)

    def resetShift(self):
        self._controlWidget.blockSignals(True)
        self._controlWidget.reset()
        self._controlWidget.blockSignals(False)
        if self._imgA and self._imgB:
            self.setImages(imgA=self._imgA, imgB=self._imgB)


class _AxisRead(qt.QWidget):
    """Widget to select a position value from a file"""

    sigFileChanged = qt.Signal(str)

    def __init__(self, parent, axis=None):
        qt.QWidget.__init__(self, parent)
        self._axis = None
        if axis:
            self.setAxis(axis)
        self.setLayout(qt.QHBoxLayout())

        self.layout().addWidget(qt.QLabel('File', parent=self))
        self._filePathQLE = qt.QLineEdit('', parent=self)
        self.layout().addWidget(self._filePathQLE)
        self._fileSelPB = qt.QPushButton('select', parent=self)
        self.layout().addWidget(self._fileSelPB)

        # connect signal / slot
        self._fileSelPB.pressed.connect(self._selectFile)
        self._filePathQLE.textChanged.connect(self._fileChanged)

    def setAxis(self, axis):
        assert isinstance(axis, QAxisRP)
        self._axis = axis

    def _selectFile(self):
        dialog = qt.QFileDialog(self)
        dialog.setFileMode(qt.QFileDialog.ExistingFile)

        if not dialog.exec_():
            dialog.close()
            return

        _file_path = dialog.selectedFiles()[0]
        _logger.info(
            'user select file %s for reading position value' % _file_path)
        self._filePathQLE.setText(dialog.selectedFiles()[0])

    def _fileChanged(self, file_path):
        """callback when the line edit (containing the file path) changed"""
        if self._axis and os.path.isfile(file_path):
            self._axis.set_position_frm_par_file(file_path, force=True)


class _AxisManual(qt.QWidget):
    """
    Widget to define the shift to apply on an image
    """

    sigShiftChanged = qt.Signal(float, float)
    """Signal emitted when requested shift changed. Parameter is x, y"""

    sigModeLockChanged = qt.Signal(bool)
    """Signal emitted when the mode is lock or unlock"""

    def __init__(self, parent, reconsParams):
        assert isinstance(reconsParams, QAxisRP)
        qt.QWidget.__init__(self, parent)
        self._xShift = 0
        self._yShift = 0
        self._recons_params = reconsParams or QAxisRP()
        self._axis = None

        self.setLayout(qt.QVBoxLayout())

        self._manualSelectionWidget = _AxisManualSelection(parent=self,
                                                           shift_mode=ShiftMode.x_only)
        self._manualSelectionWidget.layout().setContentsMargins(0, 0, 0, 0)

        self._readFileSelWidget = _AxisRead(parent=self)
        self._readFileSelWidget.layout().setContentsMargins(0, 0, 0, 0)

        self._displacementSelector = self._manualSelectionWidget._displacementSelector
        self._shiftControl = self._manualSelectionWidget._shiftControl
        self._roiControl = self._manualSelectionWidget._roiControl

        self._mainWidget = AxisTabWidget(parent=self,
                                         mode_dependant_widget=self._manualSelectionWidget,
                                         read_file_sel_widget=self._readFileSelWidget,
                                         recons_params=self._recons_params)

        self.layout().addWidget(self._mainWidget)

        # signal / slot connection
        self._shiftControl.sigShiftLeft.connect(self._incrementLeftShift)
        self._shiftControl.sigShiftRight.connect(self._incrementRightShift)
        self._shiftControl.sigShiftTop.connect(self._incrementTopShift)
        self._shiftControl.sigShiftBottom.connect(self._incrementBottomShift)
        self._shiftControl.sigReset.connect(self._resetShift)
        self._shiftControl.sigShiftChanged.connect(self._setShiftAndSignal)
        self._mainWidget.sigLockModeChanged.connect(self._modeLockChanged)

        # expose API
        self.getShiftStep = self._displacementSelector.getShiftStep
        self.setShiftStep = self._displacementSelector.setShiftStep
        self.sigRoiChanged = self._roiControl.sigRoiChanged
        self.sigAuto = self._shiftControl.sigAuto
        self.getROI = self._roiControl.getROI
        self.getMode = self._mainWidget.getMode
        self.setScan = self._mainWidget.setScan
        self.sigModeChanged = self._mainWidget.sigModeChanged
        self.isModeLock = self._mainWidget.isModeLock
        self.setModeLock = self._mainWidget.setModeLock

        # set up interface
        self.setAxis(self._recons_params)

    def _incrementLeftShift(self):
        self._incrementShift('left')

    def _incrementRightShift(self):
        self._incrementShift('right')

    def _incrementTopShift(self):
        self._incrementShift('top')

    def _incrementBottomShift(self):
        self._incrementShift('bottom')

    def _setShiftAndSignal(self, x, y):
        self.setShift(x, y)
        self._shiftControl._updateShiftInfo(x=x, y=y)
        self.sigShiftChanged.emit(x, y)

    def setAxis(self, axis):
        assert isinstance(axis, QAxisRP)
        old = self.blockSignals(True)
        if self._axis:
            self._axis.sigChanged.disconnect(self._updateAxisView)
        self._axis = axis
        self.setXShift(self._axis.value)
        self._mainWidget.setAxisParams(self._axis)
        self._readFileSelWidget.setAxis(self._axis)
        # self._positionInfoWidget.setAxis(self._axis)
        self._updateAxisView()
        self._axis.sigChanged.connect(self._updateAxisView)
        self.blockSignals(old)

    def _modeLockChanged(self, lock):
        self.sigModeLockChanged.emit(lock)

    def setMode(self, mode):
        self._axis.mode = mode
        self._updateAxisView()

    def _updateAxisView(self):
        self._axis.blockSignals(True)
        if self._axis.value not in (None, '...'):
            self.setXShift(self._axis.value)
            #self._positionInfoWidget._updatePosition()
        self._axis.blockSignals(False)

        self._manualSelectionWidget.setVisible(
            self._axis.mode is AxisMode.manual)
        self._readFileSelWidget.setVisible(self._axis.mode is AxisMode.read)

    def getAxis(self):
        return self._axis

    def _incrementShift(self, direction):
        assert direction in ('left', 'right', 'top', 'bottom')
        if direction == 'left':
            self.setXShift(self._xShift - self.getShiftStep())
        elif direction == 'right':
            self.setXShift(self._xShift + self.getShiftStep())
        elif direction == 'top':
            self.setYShift(self._yShift + self.getShiftStep())
        else:
            self.setYShift(self._yShift - self.getShiftStep())

        self._shiftControl._updateShiftInfo(x=self._xShift, y=self._yShift)
        self.sigShiftChanged.emit(self._xShift, self._yShift)

    def _resetShift(self):
        old = self._axis.blockSignals(True)
        self.setXShift(0)
        self.setYShift(0)
        self._shiftControl._updateShiftInfo(x=self._xShift, y=self._yShift)
        self._axis.blockSignals(old)

        self.sigShiftChanged.emit(self._xShift, self._yShift)

    def getXShift(self):
        if self._xShift == '...':
            return 0
        return self._xShift

    def getYShift(self):
        if self._yShift == '...':
            return 0
        return self._yShift

    def setXShift(self, x: float):
        self.setShift(x=x, y=self._yShift)

    def setYShift(self, y):
        self.setShift(x=self._xShift, y=y)

    def setShift(self, x, y):
        if x == self._xShift and y == self._yShift:
            return
        self._xShift = x if x is not None else 0.0
        self._yShift = y if y is not None else 0.0
        if self._axis:
            old = self._axis.blockSignals(True)
            self._axis.value = x
            self._axis.blockSignals(old)
        self._shiftControl._updateShiftInfo(x=self._xShift, y=self._yShift)
        self.sigShiftChanged.emit(self._xShift, self._yShift)

    def reset(self):
        self._xShift = 0
        self._yShift = 0
        self.sigShiftChanged.emit(self._xShift, self._yShift)

    def setLocked(self, locked):
        self._mainWidget.setEnabled(not locked)


class _AxisManualSelection(qt.QWidget):
    def __init__(self, parent, shift_mode):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())
        self._displacementSelector = _DisplacementSelector(parent=self)
        self.layout().addWidget(self._displacementSelector)

        self._shiftControl = _ShiftControl(parent=self, shift_mode=shift_mode)
        self.layout().addWidget(self._shiftControl)

        self._roiControl = _ROIControl(parent=self)
        self.layout().addWidget(self._roiControl)


class _ROIControl(qt.QGroupBox):
    """
    Widget used to define the ROI on images to compare
    """
    sigRoiChanged = qt.Signal(object)
    """Signal emitted when the ROI changed"""

    def __init__(self, parent):
        qt.QGroupBox.__init__(self, 'ROI selection', parent)
        self.setLayout(qt.QVBoxLayout())

        self._buttonGrp = qt.QButtonGroup(parent=self)
        self._buttonGrp.setExclusive(True)

        self._roiWidget = qt.QWidget(parent=self)
        self._roiWidget.setLayout(qt.QHBoxLayout())
        self._roiWidget.layout().setContentsMargins(0, 0, 0, 0)
        self._roiButton = qt.QRadioButton('ROI', parent=self._roiWidget)
        self._roiWidget.layout().addWidget(self._roiButton)
        self._buttonGrp.addButton(self._roiButton)
        self._roiDefinition = _ROIDefinition(parent=self)
        self._roiWidget.layout().addWidget(self._roiDefinition)
        self.layout().addWidget(self._roiWidget)

        self._fullImgButton = qt.QRadioButton('full image', parent=self)
        self._buttonGrp.addButton(self._fullImgButton)
        self.layout().addWidget(self._fullImgButton)

        # connect signal / Slot
        self._roiButton.toggled.connect(self._roiDefinition.setEnabled)

        # expose API
        self.sigRoiChanged = self._roiDefinition.sigRoiChanged
        self.getROI = self._roiDefinition.getROI
        self.setLimits = self._roiDefinition.setLimits

        # setup for full image
        self._roiButton.setChecked(True)


class _ROIDefinition(qt.QWidget):
    """
    Widget used to define ROI width and height.

    :note: emit ROI == None if setDisabled
    """
    sigRoiChanged = qt.Signal(object)
    """Signal emitted when the ROI changed"""

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QGridLayout())

        self.layout().addWidget(qt.QLabel('width', parent=self), 0, 0)
        self._widthSB = qt.QSpinBox(parent=self)
        self._widthSB.setSingleStep(2)
        self._widthSB.setMaximum(10000)
        self._widthSB.setSuffix(" px")
        self.layout().addWidget(self._widthSB, 0, 1)
        self.layout().addWidget(qt.QLabel('height', parent=self), 1, 0)
        self._heightSB = qt.QSpinBox(parent=self)
        self._heightSB.setSingleStep(2)
        self._heightSB.setSuffix(" px")
        self._heightSB.setMaximum(10000)
        self.layout().addWidget(self._heightSB, 1, 1)

        # Signal / Slot connection
        self._widthSB.editingFinished.connect(self.__roiChanged)
        self._heightSB.editingFinished.connect(self.__roiChanged)

    def __roiChanged(self, *args, **kwargs):
        self.sigRoiChanged.emit(self.getROI())

    def setLimits(self, width, height):
        """

        :param int x: width maximum value
        :param int height: height maximum value
        """
        for spinButton in (self._widthSB, self._heightSB):
            spinButton.blockSignals(True)
        assert type(width) is int
        assert type(height) is int
        valueChanged = False
        if self._widthSB.value() > width:
            self._widthSB.setValue(width)
            valueChanged = True
        if self._heightSB.value() > height:
            self._heightSB.setValue(height)
            valueChanged = True

        # if this is the first limit definition, propose default width and
        # height
        if self._widthSB.value() == 0:
            self._widthSB.setValue(min(256, width))
            valueChanged = True
        if self._heightSB.value() == 0:
            self._heightSB.setValue(min(256, height))
            valueChanged = True

        # define minimum / maximum
        self._widthSB.setRange(1, width)
        self._heightSB.setRange(1, height)
        for spinButton in (self._widthSB, self._heightSB):
            spinButton.blockSignals(False)
        if valueChanged is True:
            self.__roiChanged()

    def getROI(self):
        """

        :return: (width, height) or None
        :rtype: Union[None, tuple]
        """
        if self.isEnabled():
            return (self._widthSB.value(), self._heightSB.value())
        else:
            return None

    def setEnabled(self, *arg, **kwargs):
        qt.QWidget.setEnabled(self, *arg, **kwargs)
        self.__roiChanged()


@enum.unique
class ShiftMode(enum.Enum):
    x_only = 0
    y_only = 1
    x_and_y = 2


class _ShiftControl(qt.QWidget):
    """
    Widget to control the shift step we want to apply
    """
    sigShiftLeft = qt.Signal()
    """Signal emitted when the left button is activated"""
    sigShiftRight = qt.Signal()
    """Signal emitted when the right button is activated"""
    sigShiftTop = qt.Signal()
    """Signal emitted when the top button is activated"""
    sigShiftBottom = qt.Signal()
    """Signal emitted when the bottom button is activated"""
    sigReset = qt.Signal()
    """Signal emitted when the reset button is activated"""
    sigAuto = qt.Signal()
    """Signal emitted when the auto button is activated"""
    sigShiftChanged = qt.Signal(float, float)
    """Signal emitted ony when xLE and yLE edition is finished"""

    def __init__(self, parent, shift_mode):
        """

        :param parent: qt.QWidget
        :param ShiftMode shift_mode: what are the shift we want to control
        """
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QGridLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)

        self._leftButton = qt.QPushButton('left', parent=self)
        self.layout().addWidget(self._leftButton, 1, 0)

        self._rightButton = qt.QPushButton('right', parent=self)
        self.layout().addWidget(self._rightButton, 1, 3)

        self._topButton = qt.QPushButton('top', parent=self)
        self.layout().addWidget(self._topButton, 0, 1)

        self._bottomButton = qt.QPushButton('bottom', parent=self)
        self.layout().addWidget(self._bottomButton, 2, 1)

        self._resetButton = qt.QPushButton('reset', parent=self)
        self.layout().addWidget(self._resetButton, 3, 2, 3, 4)

        self._autoButton = qt.QPushButton('auto', parent=self)
        self.layout().addWidget(self._autoButton, 3, 0, 3, 2)
        self._autoButton.hide()

        self._shiftInfo = _ShiftInformation(parent=self)
        self.layout().addWidget(self._shiftInfo, 1, 1)
        self._shiftInfo._updateShiftInfo(x=0.0, y=0.0)

        # Signal / Slot connection
        self._leftButton.pressed.connect(self.sigShiftLeft.emit)
        self._rightButton.pressed.connect(self.sigShiftRight.emit)
        self._topButton.pressed.connect(self.sigShiftTop.emit)
        self._bottomButton.pressed.connect(self.sigShiftBottom.emit)
        self._resetButton.pressed.connect(self.sigReset.emit)
        self._autoButton.pressed.connect(self.sigAuto.emit)
        self._shiftInfo.sigShiftChanged.connect(self.sigShiftChanged.emit)

        # expose API
        self._updateShiftInfo = self._shiftInfo._updateShiftInfo

        self.setShiftMode(shift_mode)

    def setShiftMode(self, shift_mode):
        show_x_shift = shift_mode in (ShiftMode.x_only, ShiftMode.x_and_y)
        show_y_shift = shift_mode in (ShiftMode.y_only, ShiftMode.x_and_y)
        self._leftButton.setVisible(show_x_shift)
        self._rightButton.setVisible(show_x_shift)
        self._topButton.setVisible(show_y_shift)
        self._bottomButton.setVisible(show_y_shift)
        self._shiftInfo._xLE.setVisible(show_x_shift)
        self._shiftInfo._xLabel.setVisible(show_x_shift)
        self._shiftInfo._yLE.setVisible(show_y_shift)
        self._shiftInfo._yLabel.setVisible(show_y_shift)


class _ShiftInformation(qt.QWidget):
    """
    Widget displaying information about the current x and y shift.
    Both x shift and y shift are editable.
    """

    class _ShiftLineEdit(qt.QLineEdit):
        def __init__(self, *args, **kwargs):
            qt.QLineEdit.__init__(self, *args, **kwargs)
            validator = qt.QDoubleValidator(parent=self, decimals=2)
            self.setValidator(validator)

        def sizeHint(self):
            return qt.QSize(40, 10)

    sigShiftChanged = qt.Signal(float, float)
    """Signal emitted ony when xLE and yLE edition is finished"""

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QHBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)

        self._xLabel = qt.QLabel('x=', parent=self)
        self.layout().addWidget(self._xLabel)
        self._xLE = _ShiftInformation._ShiftLineEdit('', parent=self)
        self.layout().addWidget(self._xLE)

        self._yLabel = qt.QLabel('y=', parent=self)
        self.layout().addWidget(self._yLabel)
        self._yLE = _ShiftInformation._ShiftLineEdit('', parent=self)
        self.layout().addWidget(self._yLE)

        # connect Signal / Slot
        self._xLE.editingFinished.connect(self._shiftChanged)
        self._yLE.editingFinished.connect(self._shiftChanged)

    def _updateShiftInfo(self, x, y):
        self.blockSignals(True)
        if x is None:
            x = 0.0
        if y is None:
            y = 0.0
        x_text = x
        if x_text != '...':
            x_text = '%.1f' % float(x)
        y_text = y
        if y_text != '...':
            y_text = '%.1f' % float(y)
        self._xLE.setText(x_text)
        self._yLE.setText(y_text)
        self.blockSignals(False)

    def _shiftChanged(self, *args, **kwargs):
        self.sigShiftChanged.emit(float(self._xLE.text()),
                                  float(self._yLE.text()))


class _DisplacementSelector(qt.QGroupBox):
    """
    Group box to define the displacement step value
    """

    def __init__(self, parent):
        qt.QGroupBox.__init__(self, 'shift step', parent)
        self.setLayout(qt.QVBoxLayout())
        self._buttonGrp = qt.QButtonGroup(parent=self)
        self._buttonGrp.setExclusive(True)

        self._rawButton = qt.QRadioButton('Raw (1 pixel)', parent=self)
        self.layout().addWidget(self._rawButton)
        self._buttonGrp.addButton(self._rawButton)

        self._fineButton = qt.QRadioButton('Fine (0.1 pixel)', parent=self)
        self.layout().addWidget(self._fineButton)
        self._buttonGrp.addButton(self._fineButton)

        self._manualWidget = qt.QWidget(parent=self)
        self._manualWidget.setLayout(qt.QHBoxLayout())
        self._manualWidget.layout().setContentsMargins(0, 0, 0, 0)
        self._manualWidget.layout().setSpacing(0)
        self._manualButton = qt.QRadioButton('Manual',
                                             parent=self._manualWidget)
        self._manualWidget.layout().addWidget(self._manualButton)
        self._manualLE = qt.QLineEdit('0.5', parent=self._manualWidget)
        validator = qt.QDoubleValidator(parent=self._manualLE, decimals=2)
        validator.setBottom(0.0)
        self._manualLE.setValidator(validator)
        self._manualWidget.layout().addWidget(self._manualLE)

        self.layout().addWidget(self._manualWidget)
        self._manualLE.setEnabled(False)
        self._buttonGrp.addButton(self._manualButton)

        self._rawButton.setChecked(True)

        # signal / slot connection
        self._manualButton.toggled.connect(self._manualLE.setEnabled)

    def getShiftStep(self):
        """

        :return: displacement shift defined
        :rtype: float
        """
        if self._rawButton.isChecked():
            return 1.0
        elif self._fineButton.isChecked():
            return 0.1
        else:
            return float(self._manualLE.text())

    def setShiftStep(self, value):
        """

        :param float value: shift step
        """
        assert type(value) is float
        self._manualButton.setChecked(True)
        self._manualLE.setText(str(value))


class _AxisOptionsWidget(qt.QWidget):
    """GUI to tune the axis algorithm"""
    def __init__(self, parent, axis):
        qt.QWidget.__init__(self, parent=parent)
        assert isinstance(axis, QAxisRP)
        self._axis = axis
        self.setLayout(qt.QVBoxLayout())

        # define common options
        self._commonOpts = qt.QWidget(parent=self)
        self._commonOpts.setLayout(qt.QFormLayout())

        self._qcbDataMode = qt.QComboBox(parent=self)
        for data_mode in AxisCalculationInput:
            # paganin is not managed for sinogram
            self._qcbDataMode.addItem(data_mode.name(), data_mode)
        # for now not handle
        # self._commonOpts.layout().addRow('data mode', self._qcbDataMode)
        self._qcbDataMode.hide()

        # add scale option
        self._scaleOpt = qt.QCheckBox(parent=self)
        self._commonOpts.layout().addRow('scale the two images', self._scaleOpt)
        self.layout().addWidget(self._commonOpts)

        # add near options
        self._nearOpts = _AxisNearOptsWidget(parent=self, axis=self._axis)
        self.layout().addWidget(self._nearOpts)

        # set up
        self.setCalculationInputType(self._axis.calculation_input_type)

        # connect signal / slot
        self._scaleOpt.toggled.connect(self._updateScaleOpt)
        self._qcbDataMode.currentIndexChanged.connect(self._updateInputType)
        self._axis.sigChanged.connect(self._updateMode)

    def setMode(self, mode):
        mode_ = AxisMode.from_value(mode)
        self._nearOpts.setVisible(mode_ == AxisMode.near)

    def _updateMode(self):
        old = self.blockSignals(True)
        self._nearOpts.setVisible(self._axis.mode == AxisMode.near)
        index = self._qcbDataMode.findText(self._axis.calculation_input_type.name())
        if index >= 0:
            self._qcbDataMode.setCurrentIndex(index)
        self.blockSignals(old)

    def _updateScaleOpt(self, *arg, **kwargs):
        old = self._axis.blockSignals(True)
        self._axis.scale_img2_to_img1 = self.isImageScaled()
        self._axis.blockSignals(old)

    def isImageScaled(self):
        return self._scaleOpt.isChecked()

    def _updateInputType(self, *arg, **kwargs):
        self._axis.calculation_input_type = self.getCalulationInputType()

    def getCalulationInputType(self, *arg, **kwargs):
        return AxisCalculationInput.from_value(self._qcbDataMode.currentText())

    def setCalculationInputType(self, calculation_type):
        calculation_type = AxisCalculationInput.from_value(calculation_type)
        index_dm = self._qcbDataMode.findText(calculation_type.name())
        self._qcbDataMode.setCurrentIndex(index_dm)

    def setAxisParams(self, axis):
        self._nearOpts.setAxisParams(axis=axis)
        self._axis = axis
        old = self.blockSignals(True)
        self._scaleOpt.setChecked(axis.scale_img2_to_img1)
        index = self._qcbDataMode.findText(axis.calculation_input_type.name())
        self._qcbDataMode.setCurrentIndex(index)
        self.blockSignals(old)


class _AxisNearOptsWidget(qt.QWidget):
    """GUI dedicated to the neat option"""
    def __init__(self, parent, axis):
        qt.QWidget.__init__(self, parent=parent)
        assert isinstance(axis, QAxisRP)
        self._axis = axis

        self.setLayout(qt.QFormLayout())

        self._stdMaxOpt = qt.QCheckBox(parent=self)
        self.layout().addRow('look at max standard deviation', self._stdMaxOpt)

        self._nearWX = qt.QSpinBox(parent=self)
        self._nearWX.setMinimum(1)
        self._nearWX.setValue(5)
        self.layout().addRow('window size', self._nearWX)

        self._fineStepX = qt.QDoubleSpinBox(parent=self)
        self._fineStepX.setMinimum(0.05)
        self._fineStepX.setSingleStep(0.05)
        self._fineStepX.setMaximum(1.0)
        self.layout().addRow('fine step x', self._fineStepX)

        # connect signal / Slot
        self._stdMaxOpt.toggled.connect(self._lookforStxMaxChanged)
        self._nearWX.valueChanged.connect(self._windowSizeChanged)
        self._fineStepX.valueChanged.connect(self._fineStepXChanged)

    def _lookforStxMaxChanged(self, *args, **kwargs):
        self._axis.look_at_stdmax = self.isLookAtStdMax()

    def isLookAtStdMax(self):
        """

        :return: is the option for looking at max standard deviation is
                 activated
        :rtype: bool
        """
        return self._stdMaxOpt.isChecked()

    def _windowSizeChanged(self, *args, **kwargs):
        self._axis.near_wx = self.getWindowSize()

    def getWindowSize(self):
        """

        :return: window size for near search
        :rtype: int
        """
        return self._nearWX.value()

    def _fineStepXChanged(self, *args, **kwargs):
        self._axis.fine_step_x = self.getFineStepX()

    def getFineStepX(self):
        """

        :return: fine step x for near calculation
        :rtype: float
        """
        return self._fineStepX.value()

    def setAxisParams(self, axis):
        """

        :param axis: axis to edit
        :type: AxisRP
        """
        old = self.blockSignals(True)
        self._axis = axis
        self._stdMaxOpt.setChecked(axis.look_at_stdmax)
        self._nearWX.setValue(axis.near_wx)
        self._fineStepX.setValue(axis.fine_step_x)
        self.blockSignals(old)


class AxisTabWidget(qt.QTabWidget):
    """
    TabWidget containing all the information to edit the AXIS parameters
    """
    sigLockModeChanged = qt.Signal(bool)
    """signal emitted when the mode lock change"""

    def __init__(self, recons_params, parent=None, mode_dependant_widget=None,
                 read_file_sel_widget=None):
        """

        :param recons_params: reconstruction parameters edited by the widget
        :type: QAxisRP
        :param mode_dependant_widget: widget used for manual selection of the
                                      axis
        :type mode_dependant_widget: Union[None, `._AxisManualSelection`]
        :param read_file_sel_widget: widget used to select a par file containing
                                     the axis position
        :type read_file_sel_widget: Union[None, `._AxisRead`]
        """
        qt.QTabWidget.__init__(self, parent)
        assert recons_params is not None
        # first tab 'calculation widget'
        self._calculationWidget = _CalculationWidget(parent=self, axis_params=recons_params)

        # second tab: options
        self._optionsWidget = _AxisOptionsWidget(parent=self, axis=recons_params)
        self._inputWidget = _InputWidget(parent=self,
                                         axis_params=recons_params)

        if mode_dependant_widget:
            self._calculationWidget.layout().addWidget(mode_dependant_widget)

        if read_file_sel_widget:
            self._calculationWidget.layout().addWidget(read_file_sel_widget)

        for widget in self._calculationWidget, self._optionsWidget:
            spacer = qt.QWidget(self)
            spacer.setSizePolicy(qt.QSizePolicy.Minimum,
                                 qt.QSizePolicy.Expanding)
            widget.layout().addWidget(spacer)

        self._optionsSA = qt.QScrollArea(parent=self)
        self._optionsSA.setWidget(self._optionsWidget)
        self.addTab(self._calculationWidget, 'calculation')
        self.addTab(self._optionsSA, 'options')
        self.addTab(self._inputWidget, 'input')

        self.setAxisParams(recons_params)
        # set up
        self._optionsWidget.setMode(self.getMode())

        # expose API
        self.sigModeChanged = self._calculationWidget.sigModeChanged
        self.isModeLock = self._calculationWidget.isModeLock
        self.setModeLock = self._calculationWidget.setModeLock

        # connect signal / slot
        self._calculationWidget.sigLockModeChanged.connect(self._propagateSigLockModeChanged)

    def _propagateSigLockModeChanged(self, lock):
        self.sigLockModeChanged.emit(lock)

    def getMode(self):
        """Return algorithm to use for axis calculation"""
        return self._calculationWidget.getMode()

    def setScan(self, scan):
        self._inputWidget.setUrls(scan=scan)
        if scan is not None:
            self._inputWidget.setScanRange(scan.scan_range)

    def setAxisParams(self, axis):
        old = self.blockSignals(True)
        self._calculationWidget.setAxisParams(axis)
        self._optionsWidget.setAxisParams(axis)
        self._inputWidget.setAxisParams(axis)
        self.blockSignals(old)


class _CalculationWidget(qt.QWidget):
    """Main widget to select the algorithm to use for COR calculation"""

    sigModeChanged = qt.Signal(str)
    """signal emitted when the algorithm for computing COR changed"""

    sigLockModeChanged = qt.Signal(bool)
    """signal emitted when the mode has been lock or unlock"""

    def __init__(self, parent, axis_params):
        assert isinstance(axis_params, QAxisRP)
        qt.QWidget.__init__(self, parent)
        self._axis_params = None
        self.setLayout(qt.QVBoxLayout())

        self._modeWidget = qt.QWidget(parent=self)
        self._modeWidget.setLayout(qt.QHBoxLayout())
        self.layout().addWidget(self._modeWidget)

        self.__rotAxisSelLabel = qt.QLabel('algorithm to compute cor')
        self._modeWidget.layout().addWidget(self.__rotAxisSelLabel)
        self._qcbPosition = qt.QComboBox(self)
        self._qcbPosition.addItem(AxisMode.accurate.value)
        self._qcbPosition.addItem(AxisMode.manual.value)
        # self._qcbPosition.addItem(AxisMode.near.value)
        self._qcbPosition.addItem(AxisMode.read.value)
        self._modeWidget.layout().addWidget(self._qcbPosition)

        # LineEdit position value
        self._qleNearPosQLE = qt.QLineEdit('0', parent=self)
        self._modeWidget.layout().addWidget(self._qleNearPosQLE)
        validator = qt.QDoubleValidator(parent=self)
        self._qleNearPosQLE.setValidator(validator)
        self._qleNearPosQLE.hide()

        # padding mode option
        self._padding_widget = qt.QGroupBox('padding mode')
        self._padding_widget.setCheckable(True)
        self.layout().addWidget(self._padding_widget)
        self._padding_widget.setLayout(qt.QHBoxLayout())

        self._qbPaddingMode = qt.QComboBox(self._padding_widget)
        for _mode in ('constant', 'edge', 'linear_ramp', 'maximum', 'mean',
                      'median', 'minimum', 'reflect', 'symmetric', 'wrap'):
            self._qbPaddingMode.addItem(_mode)

        self._padding_widget.layout().addWidget(self._qbPaddingMode)
        self._padding_widget

        # method lock button
        self._lockMethodPB = PadlockButton(parent=self._modeWidget)
        self._lockMethodPB.setToolTip('Lock the method to compute the cor. \n'
                                      'This will automatically call the '
                                      'defined algorithm each time a scan is received.')
        self._modeWidget.layout().addWidget(self._lockMethodPB)

        # connect signal / slot
        self._qcbPosition.currentIndexChanged.connect(self._modeChanged)
        self._qleNearPosQLE.editingFinished.connect(self._nearValueChanged)
        self._lockMethodPB.sigLockChanged.connect(self.lockMode)
        self._qbPaddingMode.currentIndexChanged.connect(self._paddingModeChanged)

        # set up interface
        self.setAxisParams(axis_params)

    def _modeChanged(self, *args, **kwargs):
        self._qleNearPosQLE.setVisible(self.getMode() == AxisMode.near)
        self._padding_widget.setVisible(self.getMode() == AxisMode.accurate)
        if self.getMode() in (AxisMode.near, AxisMode.accurate):
            self._lockMethodPB.setVisible(True)
        else:
            self._lockMethodPB.setVisible(False)
            self.lockMode(False)

        self.sigModeChanged.emit(self.getMode().value)

    def isModeLock(self):
        return self._lockMethodPB.isLocked()

    def setModeLock(self, mode=None):
        """set a specific mode and lock it.

        :param mode: mode to lock. If None then keep the current mode
        """
        if mode is not None:
            mode = AxisMode.from_value(mode)
        if mode is None and self.getMode() not in (AxisMode.accurate, AxisMode.near):
            raise ValueError('Unable to lock the current mode is not an automatic algorithm')
        elif mode != self.getMode() and mode not in (AxisMode.accurate, AxisMode.near):
            raise ValueError('Unable to lock %s this is not a lockable mode')

        if mode is not None:
            self.setMode(mode)
        if not self._lockMethodPB.isLocked():
            old = self._lockMethodPB.blockSignals(True)
            self._lockMethodPB.setLock(True)
            self._lockMethodPB.blockSignals(old)
        self.lockMode(True)

    def lockMode(self, lock):
        old = self._lockMethodPB.blockSignals(True)
        self._lockMethodPB.setLock(lock)
        for widget in (self._qcbPosition, self._qleNearPosQLE, ):
            widget.setEnabled(not lock)
        self._lockMethodPB.blockSignals(old)
        self.sigLockModeChanged.emit(lock)

    def getMode(self):
        """Return algorithm to use for axis calculation"""
        return AxisMode.from_value(self._qcbPosition.currentText())


    def setMode(self, mode):
        index = self._qcbPosition.findText(mode.value)
        if index >= 0:
            self._qcbPosition.setCurrentIndex(index)

    def _nearValueChanged(self, *args, **kwargs):
        self._axis_params.near_position = self.getNearPosition()

    def getNearPosition(self):
        return float(self._qleNearPosQLE.text())

    def setNearPosition(self, position):
        self._qleNearPosQLE.setText(str(position))

    def _paddingModeChanged(self, *args, **kwargs):
        self._axis_params.padding_mode = self.getPaddingMode()

    def getPaddingMode(self):
        if self._padding_widget.isChecked():
            return self._qbPaddingMode.currentText()
        else:
            return None

    def setPaddingMode(self, mode):
        index = self._qbPaddingMode.findText(mode.value)
        if index >= 0:
            self._qbPaddingMode.setCurrentIndex(index)

    def setAxisParams(self, axis):
        old = self.blockSignals(True)
        if self._axis_params is not None:
            self._axis_params.sigChanged.disconnect(self._axis_params_changed)
        self._axis_params = axis
        self._axis_params.sigChanged.connect(self._axis_params_changed)
        self._axis_params_changed()
        self.blockSignals(old)

    def _axis_params_changed(self, *args, **kwargs):
        self.setMode(self._axis_params.mode)
        self.setNearPosition(self._axis_params.near_position)


class _InputWidget(qt.QWidget):
    """Widget used to define the radio to use for axis calculation from
    radios"""
    def __init__(self, parent=None, axis_params=None):
        assert isinstance(axis_params, QAxisRP)
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())

        self._angleModeWidget = _AngleModeGroupBox(parent=self, axis_params=axis_params)
        self.layout().addWidget(self._angleModeWidget)
        self._axisRadioUrlsWidget = _AxisUrlWidget(parent=self, axis_params=axis_params)
        self.layout().addWidget(self._axisRadioUrlsWidget)
        self._axis_params = axis_params

        # connect signal /slot
        self._axisRadioUrlsWidget.setEnabled(self._angleModeWidget._qrbCOR_manual.isChecked())

        # signal / slot connection
        self._angleModeWidget._qrbCOR_manual.toggled.connect(self._axisRadioUrlsWidget.setEnabled)

        # expose API
        self.setUrls = self._axisRadioUrlsWidget.setUrls
        self.setScanRange = self._angleModeWidget.setScanRange

    def setAxisParams(self, axis_params):
        self._angleModeWidget.setAxisParams(axis_params)
        self._axisRadioUrlsWidget.setAxisParams(axis_params)
        self._axis_params = axis_params


class _AngleModeGroupBox(qt.QGroupBox):
    """Group box to select the angle to used for cor calculation
    (0-180, 90-270 or manual)"""
    def __init__(self, parent=None, axis_params=None):
        assert isinstance(axis_params, QAxisRP)
        qt.QGroupBox.__init__(self, parent=parent, title='Angles to use for axis calculation')
        self._corButtonsGps = qt.QButtonGroup(parent=self)
        self._corButtonsGps.setExclusive(True)
        self.setLayout(qt.QGridLayout())
        self._qrbCOR_0_180 = qt.QRadioButton('0-180', parent=self)
        self.layout().addWidget(self._qrbCOR_0_180, 0, 0)
        self._qrbCOR_90_270 = qt.QRadioButton('90-270',
                                              parent=self)
        self._qrbCOR_90_270.setToolTip('pick radio closest to angles 90° and '
                                       '270°. If disable mean that the scan '
                                       'range is 180°')
        self.layout().addWidget(self._qrbCOR_90_270, 0, 1)
        self._qrbCOR_manual = qt.QRadioButton('manual selection',
                                              parent=self)
        self._qrbCOR_manual.setVisible(False)
        self.layout().addWidget(self._qrbCOR_manual, 0, 2)
        # add all button to the button group
        for b in (self._qrbCOR_0_180, self._qrbCOR_90_270,
                  self._qrbCOR_manual):
            self._corButtonsGps.addButton(b)

        self.setAxisParams(axis_params)

        # connect signal / Slot
        self._corButtonsGps.buttonClicked.connect(self._angleModeChanged)

    def setScanRange(self, scanRange):
        if scanRange == 180:
            self._qrbCOR_90_270.setEnabled(False)
            if self._qrbCOR_90_270.isChecked():
                self._qrbCOR_0_180.setChecked(True)

    def setAngleMode(self, mode):
        """"

        :param mode: mode to use (can be manual , 90-270 or 0-180)
        """
        assert isinstance(mode, CorAngleMode)
        if mode == CorAngleMode.use_0_180:
            self._qrbCOR_0_180.setChecked(True)
        elif mode == CorAngleMode.use_90_270:
            self._qrbCOR_90_270.setChecked(True)
        else:
            self._qrbCOR_manual.setChecked(True)

    def getAngleMode(self):
        """

        :return: the angle to use for the axis calculation
        :rtype: CorAngleMode
        """
        if self._qrbCOR_90_270.isChecked():
            return CorAngleMode.use_90_270
        elif self._qrbCOR_0_180.isChecked():
            return CorAngleMode.use_0_180
        else:
            return CorAngleMode.manual_selection

    def setAxisParams(self, axis_params):
        old = self.blockSignals(True)
        self._axis_params = axis_params
        # set up
        self.setAngleMode(axis_params.angle_mode)
        self.blockSignals(old)

    def _angleModeChanged(self, *args, **kwargs):
        self._axis_params.angle_mode = self.getAngleMode()


class _AxisUrlWidget(qt.QWidget):
    """
    Widget used to show url the user can select for computing axis position
    """

    _RADIO_1_NAME = 'radio 1'

    _RADIO_2_NAME = 'radio 2'

    def __init__(self, parent=None, axis_params=None):
        qt.QWidget.__init__(self, parent=parent)
        # ned to be shared with the other widget as storing some values in it
        assert isinstance(axis_params, QAxisRP)
        self._recons_params = axis_params
        self._scan = None
        """Scan currently treated"""
        self._scan_proj_urls = None
        """known scan projections files"""

        self.setLayout(qt.QGridLayout())
        self.layout().addWidget(qt.QLabel(self._RADIO_1_NAME, parent=self), 0, 0)
        self.layout().addWidget(qt.QLabel(self._RADIO_2_NAME, parent=self), 1, 0)

        self._url1QLE = qt.QLineEdit(parent=self)
        self.layout().addWidget(self._url1QLE, 0, 1, 1, 2)
        self._url2QLE = qt.QLineEdit(parent=self)
        self.layout().addWidget(self._url2QLE, 1, 1, 1, 2)

        self._selectButton = qt.QPushButton('select', parent=self)
        self.layout().addWidget(self._selectButton, 2, 2)
        self._selectButton.hide()

        spacer = qt.QWidget(parent=self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(spacer, 3, 0)

        columns = OrderedDict([
            (self._RADIO_1_NAME, ColumnMode.SINGLE),
            (self._RADIO_2_NAME, ColumnMode.SINGLE),
        ])
        self._urlSelectionTable = UrlSelectionDialog(columns=columns, parent=None)

        # signal / slot connection
        self._selectButton.clicked.connect(self._selectUrls)
        self._url1QLE.editingFinished.connect(self._url1Updated)
        self._url2QLE.editingFinished.connect(self._url2Updated)

    def _url1Updated(self):
        from silx.io.url import DataUrl
        try:
            url = DataUrl(path=self._url1QLE.text())
        except:
            _logger.info('fail to create an url out of', self._url1QLE.text())
        else:
            # axis url is specific to the scan
            self._scan.axis_params.axis_url_1 = url

    def _url2Updated(self):
        from silx.io.url import DataUrl
        try:
            url = DataUrl(path=self._url2QLE.text())
        except:
            _logger.info('fail to create an url out of', self._url2QLE.text())
        else:
            # axis url is specific to the scan
            self._scan.axis_params.axis_url_2 = url

    def _updateAxisFiles(self, scan):
        assert scan is not None
        self._scan = scan

        if scan.axis_params is not None and scan.axis_params.n_url() == 2:
            axis = scan.axis_params
            self.updateUrl1Value(prop=axis.axis_url_1.url)
            self.updateUrl2Value(prop=axis.axis_url_2.url)

    def updateUrl1Value(self, prop):
        old = self.blockSignals(True)
        self._updateAxisValue(qle=self._url1QLE, prop=prop)
        self.blockSignals(old)

    def updateUrl2Value(self, prop):
        old = self.blockSignals(True)
        self._updateAxisValue(qle=self._url2QLE, prop=prop)
        self.blockSignals(old)

    def _updateAxisValue(self, qle, prop):
        if prop is None:
            value = ''
        else:
            value = prop.path()
        qle.setText(value)

    def setUrls(self, scan):
        """
        Update the interface to fit to the given scan and urls

        :param TomoBase scan:
        """
        assert scan is not None
        if (self._scan is None or self._scan.path != scan.path):
            self._scan = scan
            self._scan_proj_urls = scan.projections
        self._updateAxisFiles(scan=scan)
        if scan.axis_params is not None:
            n_url = scan.axis_params.n_url()
            if n_url >= 1:
                url1 = scan.axis_params.axis_url_1.url
                self._urlSelectionTable.addUrl(url1)
            if n_url >= 2:
                url2 = scan.axis_params.axis_url_2.url
                self._urlSelectionTable.addUrl(url2)
        self._scan = scan

    def _selectUrls(self):
        assert self._scan
        assert self._scan.axis_params
        # update the selection table
        proj_urls = self._scan.projections
        self._urlSelectionTable.setUrls(proj_urls.values())

        axis = self._scan.axis_params

        sel_url_1 = None
        if axis.axis_url_1 is not None:
            sel_url_1 = [axis.axis_url_1.url,]
        sel_url_2 = None
        if axis.axis_url_2 is not None:
            sel_url_2 = [axis.axis_url_2.url,]

        selection = {
            self._RADIO_1_NAME: sel_url_1,
            self._RADIO_2_NAME: sel_url_2,
        }
        self._urlSelectionTable.setSelection(selection=selection)
        if self._urlSelectionTable.exec_():
            selection = self._urlSelectionTable.getSelection()

            def apply_selection(sel_name):
                sel = selection[sel_name]
                if sel is not None:
                    assert len(sel) == 1
                    sel = sel[0]
                return sel
            # url 1
            sel = apply_selection(self._RADIO_1_NAME)
            # axis url is specific to the scan
            self._scan.axis_params.axis_url_1 = sel
            self.updateUrl1Value(sel)
            # url 2
            sel = apply_selection(self._RADIO_2_NAME)
            # axis url is specific to the scan
            self._scan.axis_params.axis_url_2 = sel
            self.updateUrl2Value(sel)

    def setAxisParams(self, axis):
        self._recons_params = axis



# coding: utf-8
###########################################################################
# Copyright (C) 2016-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "21/02/2020"


from silx.gui import qt
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.process.reconstruction import nabu
from tomwer.core.settings import get_lbsram_path, isOnLbsram
from tomwer.core.utils import isLowOnMemory
from ..processingstack import FIFO, ProcessingThread
import logging
import h5py

_logger = logging.getLogger(__name__)


class NabuProcessStack(FIFO, qt.QObject):
    """Implementation of the `.AxisProcess` but having a stack for treating
    scans and making computation in threads"""
    def __init__(self, parent=None):
        qt.QObject.__init__(self, parent=parent)
        FIFO.__init__(self)
        self._dry_run = False

    def _create_processing_thread(self):
        return _ProcessingThread()

    def _process(self, scan, configuration, callback=None):
        self._scan_currently_computed = scan
        assert isinstance(scan, TomwerScanBase)
        if isOnLbsram() and isLowOnMemory(get_lbsram_path()) is True:
            # if computer is running into low memory on lbsram skip it
            mess = 'low memory, skip nabu reconstruction for', scan.path
            _logger.processSkipped(mess)
            self._end_threaded_computation()
        else:
            self._computationThread.init(scan=scan,
                                         configuration=configuration)
            self._computationThread.setDryRun(self._dry_run)
            # need to manage connect before starting it because
            self._computationThread.finished.connect(self._end_threaded_computation)
            self._computationThread.start()

    def _create_processing_thread(self) -> qt.QThread:
        return _ProcessingThread()

    def setDryRun(self, dry_run):
        self._dry_run = dry_run


class _ProcessingThread(ProcessingThread):
    """
    Thread use to execute the processing of nabu reconstruction
    """
    def __init__(self):
        ProcessingThread.__init__(self)
        self._scan = None
        self._configuration = None
        self._dry_run = False

    def setDryRun(self, dry_run):
        self._dry_run = dry_run

    def init(self, scan, configuration):
        self._scan = scan
        self._configuration = configuration

    def run(self):
        self.sigComputationStarted.emit()
        # nabu can access some data from the process file.
        nabu.run_reconstruction(config=self._configuration, scan=self._scan,
                                dry_run=self._dry_run)
        index = self._scan.pop_process_index()
        if isinstance(self._scan, EDFTomoScan):
            entry = None
        else:
            entry = self._scan.entry
        nabu.run_reconstruction(config=self._configuration, scan=self._scan,
                                dry_run=self._dry_run)

        index = self._scan.pop_process_index()
        if isinstance(self._scan, EDFTomoScan):
            entry = None
        else:
            entry = self._scan.entry
        nabu.Nabu._register_process(process_file=self._scan.process_file,
                                    entry=entry,
                                    configuration=self._configuration,
                                    results={},
                                    process=nabu.Nabu,
                                    process_index=index,
                                    overwrite=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import sys
from silx.gui import qt
import argparse
from tomwer.gui import icons
from tomwer.gui.utils.splashscreen import getMainSplashScreen
from tomwer.gui.reconstruction.nabu.nabu import NabuDialog
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomwer.core.process.reconstruction.nabu import run_reconstruction
import signal

logging.basicConfig()
_logger = logging.getLogger(__name__)


def getinputinfo():
    return "tomwer nabu [scan_path]"


def sigintHandler(*args):
    """Handler for the SIGINT signal."""
    qt.QApplication.quit()


class NabuProcessingThread(qt.QThread):
    def init(self, scan, config, dry_run=False):
        self._scan = scan
        self._config = config
        self._dry_run = dry_run

    def run(self):
        run_reconstruction(scan=self._scan, config=self._config,
                           dry_run=self._dry_run)


class NabuWindow(NabuDialog):
    def __init__(self, parent=None):
        NabuDialog.__init__(self, parent)
        self.__scan = None
        self.__processingThread = NabuProcessingThread()
        self._dry_run = False

        # connect signal / slot
        self._computePB.released.connect(self._requireComputation)

    def setScan(self, scan):
        self.__scan = scan

    def getScan(self):
        return self.__scan

    def _requireComputation(self, *arg, **kwargs):
        self.setEnabled(False)
        qt.QApplication.setOverrideCursor(qt.Qt.WaitCursor)
        scan = self.getScan()
        if scan is not None:
            self._launchComputation(scan=scan, config=self.getConfiguration())

    def _computationEnded(self):
        self.setEnabled(True)
        qt.QApplication.restoreOverrideCursor()

    def getProcessingThread(self):
        return self.__processingThread

    def _launchComputation(self, scan, config):
        # update the processing thread
        thread = self.getProcessingThread()
        thread.init(scan=scan, config=config, dry_run=self._dry_run)
        thread.finished.connect(self._computationEnded)

        # start processing
        thread.start(priority=qt.QThread.LowPriority)

    def set_dry_run(self, dry_run):
        self._dry_run = dry_run


def main(argv):
    import os
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        'scan_path',
        help='Data file to show (h5 file, edf files, spec files)')
    parser.add_argument(
        '--entry',
        default=None,
        help='An entry should be specify for hdf5 files')
    parser.add_argument(
        '--output-dir',
        help='output directory of the reconstruction(s)')
    parser.add_argument(
        '--debug',
        dest="debug",
        action="store_true",
        default=False,
        help='Set logging system in debug mode')
    parser.add_argument(
        '--dry-run',
        dest="dry_run",
        action="store_true",
        default=False,
        help='Only create the .nabu file and print the configuration used')
    options = parser.parse_args(argv[1:])

    global app  # QApplication must be global to avoid seg fault on quit
    app = qt.QApplication.instance() or qt.QApplication([])

    qt.QLocale.setDefault(qt.QLocale(qt.QLocale.English))
    qt.QLocale.setDefault(qt.QLocale.c())
    signal.signal(signal.SIGINT, sigintHandler)
    sys.excepthook = qt.exceptionHandler
    timer = qt.QTimer()
    timer.start(500)
    # Application have to wake up Python interpreter, else SIGINT is not
    # catch
    timer.timeout.connect(lambda: None)

    splash = getMainSplashScreen()
    options.scan_path = options.scan_path.rstrip(os.path.sep)
    if options.entry is None:
        scan = ScanFactory.create_scan_object(options.scan_path)
    else:
        scan = HDF5TomoScan(scan=options.scan_path, entry=options.entry)
    if scan is None:
        raise ValueError('Given scan path is not recognized as a path'
                         'containing a scan')
    widget = NabuWindow(parent=None)
    widget.setScan(scan)
    if options.output_dir is not None:
        widget.setOutputDir(options.output_dir)
    # for the application we run for the reconstruction to be finished
    # to give back hand to the user
    widget.set_dry_run(options.dry_run)
    widget.setWindowTitle('Nabu reconstruction')
    widget.setWindowIcon(icons.getQIcon('tomwer'))
    splash.finish(widget)
    widget.show()
    app.exec_()


if __name__ == '__main__':
    main(sys.argv)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "05/04/2019"


import shutil
import tempfile
import unittest
import numpy
import os
import fabio
import h5py
from tomwer.core.process.reconstruction.darkref.darkrefs import DarkRefs
from tomwer.test.utils import UtilsTest
from tomwer.core.utils.scanutils import MockEDF, MockHDF5
from ..reconstruction.axis.axis import AxisProcess
from ..reconstruction.axis.axis import compute_near
from ..reconstruction.axis.axis import compute_scan_near
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.process.reconstruction.axis.params import AxisRP
from tomwer.core.process.reconstruction.axis.mode import AxisMode
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.process.baseprocess import BaseProcess
from tomwer.core.process.reconstruction.axis.params import AxisCalculationInput
from silx.io.url import DataUrl


class _AxisDataset:
    def __init__(self, name):
        self.name = name


class TestAxisIO(unittest.TestCase):
    """Test inputs and outputs types of the handler functions"""

    @staticmethod
    def _random_calc(scan):
        return numpy.random.random()

    def setUp(self):
        self.scan_folder = tempfile.mkdtemp()

        self.scan = MockEDF.mockScan(scanID=self.scan_folder,
                                     nRadio=10,
                                     nRecons=1,
                                     nPagRecons=4,
                                     dim=10)
        self.recons_params = AxisRP()
        self.axis_process = AxisProcess(self.recons_params)

        # set the axis url to be used
        projections = self.scan.projections
        urls = list(projections.values())
        self.scan.axis_params = AxisRP()
        self.scan.axis_params.axis_url_1 = urls[0]
        self.scan.axis_params.axis_url_2 = urls[1]

        # patch the axis process
        self.axis_process.RADIO_CALCULATIONS_METHODS[AxisMode.accurate] = TestAxisIO._random_calc


    def tearDown(self):
        shutil.rmtree(self.scan_folder)

    def testInputOutput(self):
        """Test that io using TomoBase instance work"""
        # for input_type in (dict, TomoBase):
        for input_type in (TomwerScanBase,):
            for _input in AxisProcess.inputs:
                for return_dict in (True, False):
                    with self.subTest(handler=_input.handler,
                                      return_dict=return_dict,
                                      input_type=input_type):
                        input_obj = self.scan
                        if input_obj is dict:
                            input_obj = input_obj.to_dict()
                        self.axis_process._set_return_dict(return_dict)
                        out = getattr(self.axis_process, _input.handler)(input_obj)
                        if return_dict:
                            self.assertTrue(isinstance(out, dict))
                        else:
                            self.assertTrue(isinstance(out, TomwerScanBase))


@unittest.skip('Axis algorithm are pending at the moment')
class TestAxisComputation(unittest.TestCase):
    """Test functions used for computing the center of rotation"""
    def setUp(self):
        unittest.TestCase.setUp(self)
        # create a folder with an unfinished acquisition
        self.tmp_folder = tempfile.mkdtemp()

    def tearDown(self):
        # create a folder with an unfinished acquisition
        if os.path.exists(self.tmp_folder):
            shutil.rmtree(path=self.tmp_folder)
        unittest.TestCase.tearDown(self)

    def test_compute_near(self):
        """Test results of the near algorithm"""
        dataset1 = _AxisDataset('00021_PURfoam2_0N_500angles')
        dataset1.radio_1 = '00021_PURfoam2_0N_500angles3000.edf'
        dataset1.radio_2 = '00021_PURfoam2_0N_500angles3002.edf'
        dataset1.near_positions = (0.0, 1.1)
        dataset1.window_sizes = (2, 5, 10)
        dataset1.fine_steps = (0.1, )
        dataset1.look_at_stdmax = (False, )

        for dataset in (dataset1, ):
            for near_position in dataset.near_positions:
                for window_size in dataset.window_sizes:
                    for fine_step in dataset.fine_steps:
                        for look_at_stdmax in dataset.look_at_stdmax:
                            self.subTest("near cor calculation",
                                         radios=(dataset.radio_1, dataset.radio_2),
                                         near_position=near_position,
                                         look_at_stdmax=look_at_stdmax,
                                         fine_step=fine_step,
                                         window_size=window_size)

                            scan_dir = os.path.join(
                                UtilsTest.getDatasets(name='axis'),
                                dataset.name)
                            file1 = os.path.join(scan_dir, dataset.radio_1)
                            file2 = os.path.join(scan_dir, dataset.radio_2)

                            with fabio.open(file1) as _f:
                                radio1 = _f.data
                            with fabio.open(file2) as _f:
                                radio2 = _f.data
                            radio1, radio2 = self.apply_home_made_normalization(
                                radio1=radio1,
                                radio2=radio2,
                                scan_dir=scan_dir,
                                use_log=False)

                            res = compute_near(radio_1=radio1,
                                               radio_2=radio2,
                                               near_pos=near_position,
                                               window_size=window_size,
                                               fine_step_x=fine_step,
                                               look_at_stdmax=look_at_stdmax)
                            self.assertTrue(
                                numpy.isclose(res, self.th_res_near(name=dataset.name,
                                                                    near_pos=near_position,
                                                                    window_size=window_size,
                                                                    fine_step=fine_step,
                                                                    look_at_stdmax=look_at_stdmax))
                            )

    def test_img_scale(self):
        """Test the algorithm within the scale_img2_to_img1 option"""
        dataset = _AxisDataset('A2_550N_20k_')
        dataset.radio_1 = 'A2_550N_20k_2499.edf'
        dataset.radio_2 = 'A2_550N_20k_2501.edf'

        scan_dir = os.path.join(UtilsTest.getDatasets(name='axis'), dataset.name)
        scan = EDFTomoScan(scan_dir)
        scan.axis_params = AxisRP()
        scan.axis_params.axis_url_1 = DataUrl(file_path=os.path.join(scan.path, dataset.radio_1),
                                              scheme='fabio', data_slice=(0,))
        scan.axis_params.axis_url_2 = DataUrl(file_path=os.path.join(scan.path, dataset.radio_2),
                                              scheme='fabio', data_slice=(0,))
        scan.axis_params.calculation_input_type = AxisCalculationInput.emission
        scan.axis_params.scale_im2_to_im1 = True

        self.assertTrue(numpy.isclose(compute_scan_global(scan=scan), 256.0))

        scan.axis_params.mode = AxisMode.near
        scan.axis_params.fine_step_x = 0.1
        scan.axis_params.near_wx = 5
        scan.axis_params.look_at_std_max = False
        scan.axis_params.near_position = 255.5
        self.assertTrue(numpy.isclose(compute_scan_near(scan=scan), 257.8))

    def th_res_near(self, name, near_pos, window_size, fine_step, look_at_stdmax):
        if name == '00021_PURfoam2_0N_500angles':
            if near_pos == 0 and fine_step == 0.1 and look_at_stdmax:
                if window_size == 5:
                    return 2.65
                elif window_size == 10:
                    return -4.35
                elif window_size == 2:
                    return 0.65
            elif near_pos == 0 and fine_step == 0.1 and not look_at_stdmax:
                if window_size == 5:
                    return 0.3
                elif window_size == 10:
                    return -3.2
                elif window_size == 2:
                    return 0.3
            elif near_pos == 1.1 and fine_step == 0.1 and look_at_stdmax:
                if window_size == 5:
                    return 2.65
                elif window_size == 10:
                    return -3.95
                elif window_size == 2:
                    return 2.2
            elif near_pos == 1.1 and fine_step == 0.1 and not look_at_stdmax:
                if window_size == 5:
                    return 0.3
                elif window_size == 10:
                    return -3.2
                elif window_size == 2:
                    return 0.3

    def apply_home_made_normalization(self, radio1, radio2, scan_dir, use_log):
        refMat = numpy.zeros_like(radio1)
        darkMat = numpy.zeros_like(radio1)
        reflist = DarkRefs.getRefHSTFiles(scan_dir, prefix='ref')
        darklist = DarkRefs.getDarkHSTFiles(scan_dir,
                                            prefix='dark')

        for reffile in reflist:
            fd = fabio.open(reffile)
            refMat = refMat + fd.data
            fd.close()

        refMat = refMat / len(reflist)

        for darkfile in darklist:
            fd = fabio.open(darkfile)
            darkMat = darkMat + fd.data
            fd.close()

        darkMat = darkMat / len(darklist)
        refMat -= darkMat
        refMat[refMat < 1] = 1.

        radio1 = (radio1 - darkMat) / refMat
        radio2 = (radio2 - darkMat) / refMat

        if use_log is True:
            radio1 = -numpy.log(radio1)
            radio2 = -numpy.log(radio2)

        radio1[numpy.isnan(radio1)] = 0.0001
        radio2[numpy.isnan(radio2)] = 0.0001
        return radio1, radio2

    def th_res(self, name, use_log):
        if name == '00021_PURfoam2_0N_500angles':
            if use_log is True:
                return 9.040816326530603
            else:
                return -319.84375


class TestAxis(unittest.TestCase):
    """Test the axis process"""
    def setUp(self):
        self.recons_params = AxisRP()
        self.axis_process = AxisProcess(axis_params=self.recons_params)
        self.recons_params.mode = 'accurate'

    # def test_no_axis_files(self):
    #     """Test that if no projections are found for computing axis the process
    #     will return None"""
    #     scan = ScanFactory.mock_scan()
    #     self.assertEqual(self.axis_process.process(scan=scan), None)

    def test_process_saved_edf(self):
        """Test that if process is called, the tomwer.h5 file is created
        and is correctly saving information regarding the center of position
        """
        self.tempdir = tempfile.mkdtemp()
        mock = MockEDF(scan_path=self.tempdir, n_radio=10, n_ini_radio=10)
        scan = EDFTomoScan(mock.scan_path)
        self.recons_params.mode = 'near'
        self.axis_process.process(scan=scan)
        self.assertTrue(os.path.exists(scan.process_file))

        with h5py.File(scan.process_file, 'r', swmr=True) as h5f:
            self.assertTrue('entry' in h5f)
            self.assertTrue('tomwer_process_0' in h5f['entry'])
            group_axis = h5f['entry']['tomwer_process_0']
            self.assertTrue('configuration' in group_axis)
            self.assertTrue('program' in group_axis)
            self.assertTrue('results' in group_axis)
            self.assertTrue('center_of_rotation' in group_axis['results'])
            axis_value = group_axis['results']['center_of_rotation'][()]

        processes = BaseProcess.get_processes_frm_type(process_file=scan.process_file,
                                                       process_type=AxisProcess,
                                                       entry='entry')
        self.assertEqual(len(processes), 1)
        self.assertEqual(processes[0].results['center_of_rotation'], axis_value)

        # make sure the process file can be loaded by an instance of AxisRP
        new_axis_params = AxisRP()
        new_axis_params.value = -10000
        new_axis_params.set_position_frm_par_file(file_path=scan.process_file,
                                                  entry=None)
        self.assertEqual(new_axis_params.value, scan.axis_params.value)

    def test_process_saved_hdf5(self):
        """Test that if process is called, the tomwer.h5 file is created
        and is correctly saving information regarding the center of position
        """
        self.tempdir = tempfile.mkdtemp()
        dim = 10
        mock = MockHDF5(scan_path=self.tempdir, n_proj=10, n_ini_proj=10,
                        scan_range=180, dim=dim)
        mock.add_alignment_radio(index=10, angle=90)
        mock.add_alignment_radio(index=10, angle=0)
        scan = mock.scan
        self.recons_params.mode = 'near'

        # check data url take
        self.axis_process.process(scan=scan)
        # make sure center of position has been computed
        self.assertTrue(os.path.exists(scan.process_file))

        with h5py.File(scan.process_file, 'r', swmr=True) as h5f:
            self.assertTrue('entry' in h5f)
            self.assertTrue('tomwer_process_0' in h5f['entry'])
            group_axis = h5f['entry']['tomwer_process_0']
            self.assertTrue('configuration' in group_axis)
            self.assertTrue('program' in group_axis)
            self.assertTrue('results' in group_axis)
            self.assertTrue('center_of_rotation' in group_axis['results'])
            axis_value = group_axis['results']['center_of_rotation'][()]

        self.assertTrue(-dim / 2 <= axis_value <= dim / 2)
        processes = BaseProcess.get_processes_frm_type(
            process_file=scan.process_file,
            process_type=AxisProcess)
        self.assertEqual(len(processes), 1)
        self.assertEqual(processes[0].results['center_of_rotation'],
                         axis_value)

        # make sure the process file can be loaded by an instance of AxisRP
        new_axis_params = AxisRP()
        new_axis_params.value = -10000
        new_axis_params.set_position_frm_par_file(
            file_path=scan.process_file)
        self.assertEqual(new_axis_params.value, scan.axis_params.value)


class TestAxisRP(unittest.TestCase):
    """Test the class used for AxisProcess configuration"""
    def setUp(self):
        self.axis_rp = AxisRP()
        self.tmp_folder = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tmp_folder)

    def testSetPositionFromParFile(self):
        """Test that the position can be managed from a par file"""
        old_value = self.axis_rp.value

        with self.assertRaises(ValueError):
            self.axis_rp.set_position_frm_par_file('ghost_file.par', force=False)
        self.assertTrue(self.axis_rp.value == old_value)

        existing_empty_file = os.path.join(self.tmp_folder, 'par_file.par')
        with open(existing_empty_file, 'w') as _file:
            _file.write('toto')

        with self.assertRaises(KeyError):
            self.axis_rp.set_position_frm_par_file(existing_empty_file, force=False)
        self.assertTrue(self.axis_rp.value == old_value)

        valid_par_file = os.path.join(UtilsTest.getEDFDataset('scan_3_'), 'scan_3_slice.par')
        assert os.path.isfile(valid_par_file)
        self.axis_rp.set_position_frm_par_file(valid_par_file, force=False)
        self.assertTrue(self.axis_rp.value == 1024.742042)

        self.axis_rp.set_position_frm_par_file(existing_empty_file, force=True)
        self.assertTrue(self.axis_rp.value is None)

        self.axis_rp.set_position_frm_par_file('ghost_file.par', force=True)
        self.assertTrue(self.axis_rp.value is None)

        self.axis_rp.set_position_frm_par_file(valid_par_file, force=True)
        self.assertTrue(self.axis_rp.value == 1024.742042)


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestAxisIO, TestAxisComputation, TestAxis, TestAxisRP):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite



#/*##########################################################################
# Copyright (C) 2016-2017 European Synchrotron Radiation Facility
#
# This file is part of the PyMca X-ray Fluorescence Toolkit developed at
# the ESRF by the Software group.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__author__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "02/06/2017"

from silx.gui import qt
import numpy
import os
from silx.io.utils import get_data as silx_get_data
from tomwer.core.utils import ftseriesutils
import tomwer.resources
from tomwer.core.log import TomwerLogger

logger = TomwerLogger(__name__)


class ImageLoaderThread(qt.QThread):
    """Thread used to load an image"""
    IMG_NOT_FOUND = numpy.load(
            tomwer.resources._resource_filename('%s.%s' % ('imageNotFound', 'npy'),
                                                default_directory=os.path.join(
                                                        'gui', 'icons'))
    )


    def __init__(self, url):
        """

        :param index: index of the image on the stackplot
        :param filePath: filePath is the file to load on stackplot reference.
                         It can be an .edf file or a .vol file. If this is a
                         vol file then the name is given with the slice z index
                         to be loaded.
        """
        super(qt.QThread, self).__init__()
        self.data = None
        self.url = url

    def getData(self):
        if hasattr(self, 'data'):
            return self.data
        else:
            return None

    def run(self):
        if os.path.exists(self.url.file_path()) and os.path.isfile(self.url.file_path()):
            if (self.url.file_path().lower().endswith('.vol.info') or
                    self.url.file_path().lower().endswith('.vol')):
                self.data = self._loadVol()
            elif self.url.scheme() == 'numpy':
                self.data = numpy.load(self.url.file_path())
            else:
                try:
                    self.data = silx_get_data(self.url)
                except Exception as e:
                    logger.error(e)
                    logger.warning(
                        'file %s not longer exists or is empty' % self.url)
                    self.data = None
        else:
            logger.warning('file %s not longer exists or is empty' % self.url)
            self.data = self.IMG_NOT_FOUND

    def _loadVol(self):
        if self.url.file_path().lower().endswith('.vol.info'):
            infoFile = self.url.file_path()
            rawFile = self.url.file_path().replace('.vol.info', '.vol')
        else:
            assert self.url.file_path().lower().endswith('.vol')
            rawFile = self.url.file_path()
            infoFile = self.url.file_path().replace('.vol', '.vol.info')

        if not os.path.exists(rawFile):
            data = None
            mess = "Can't find raw data file %s associated with %s" % (rawFile, infoFile)
            logger.warning(mess)
        elif not os.path.exists(infoFile):
            mess = "Can't find info file %s associated with %s" % (infoFile, rawFile)
            logger.warning(mess)
            data = None
        else:
            shape = ftseriesutils._getShapeForVolFile(infoFile)
            if None in shape:
                logger.warning(
                    'Fail to retrieve data shape for %s.' % infoFile)
                data = None
            else:
                try:
                    numpy.zeros(shape)
                except MemoryError:
                    data = None
                    logger.warning('Raw file %s is to large for being '
                                   'readed %s' % rawFile)
                else:
                    data = numpy.fromfile(rawFile, dtype=numpy.float32,
                                          count=-1, sep='')
                    try:
                        data = data.reshape(shape)
                    except ValueError:
                        logger.warning('unable to fix shape for raw file %s. '
                                       'Look for information in %s'
                                       '' % (rawFile, infoFile))
                        try:
                            sqr = int(numpy.sqrt(len(data)))
                            shape = (1, sqr, sqr)
                            data = data.reshape(shape)
                        except ValueError:
                            logger.info('deduction of shape size for %s failed'
                                        % rawFile)
                            data = None
                        else:
                            logger.warning('try deducing shape size for %s '
                                           'might be an incorrect '
                                           'interpretation' % rawFile)
        if self.url.data_slice() is None:
            return data
        else:
            return data[self.url.data_slice()]



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "14/02/2017"


import logging
import unittest
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt
from tomwer.core.process.reconstruction.ftseries.params.beamgeo import BeamGeoType
from tomwer.core.process.reconstruction.darkref.params import \
    Method as dkrf_method
from tomwer.core.process.reconstruction.ftseries.params.ft import VolSelMode
from tomwer.core.process.reconstruction.ftseries.params.paganin import PaganinMode
from tomwer.core.process.reconstruction.ftseries.params.pyhst import PyhstRP
from tomwer.gui.qtapplicationmanager import QApplicationManager
from tomwer.gui.reconstruction.ftserie.reconsparamseditor import ReconsParamsEditor
from tomwer.synctools.ftseries import QReconsParams
from tomwer.test.utils import skip_gui_test

# Makes sure a QApplication exists
_qapp = QApplicationManager()

logging.disable(logging.INFO)


@unittest.skipIf(skip_gui_test(), reason='skip gui test')
class TestSimplifyH5EditorDisplay(TestCaseQt):
    """Make sure we are displayig the correct things
    """
    def setUp(self):
        super(TestSimplifyH5EditorDisplay, self).setUp()
        self.h5Editor = ReconsParamsEditor(None)
        self.h5Editor.setAttribute(qt.Qt.WA_DeleteOnClose)

        self.input = QReconsParams()

    def tearDown(self):
        self.h5Editor.close()
        del self.h5Editor
        self.qapp.processEvents()
        super(TestCaseQt, self).tearDown()

    def testPAGANIN(self):
        """ As test are 'raw' and based on the values of the FastSetupDefineGlobals
        We are checking the values of FastSetupDefineGlobals by assert
        to make the difference with the 'True' unittest
        """
        widget = self.h5Editor._PaganinWidget
        widget.load(self.input)

        # check Mode conbobox
        assert 'MODE' in self.input['PAGANIN'].all_params
        assert self.input['PAGANIN']['MODE'] == PaganinMode.off
        self.assertTrue(widget._qcbpaganin.currentIndex() == 0)
        self.assertTrue(widget._qcbpaganin.currentText() == "off")

        # test alpha/beta
        assert 'DB2' in self.input['PAGANIN'].all_params
        assert self.input['PAGANIN']['DB2'] == 100.
        self.assertTrue(widget._qleSigmaBeta2.text() == str(100.))

        # unsharp coefficient
        assert 'UNSHARP_COEFF' in self.input['PAGANIN'].all_params
        assert self.input['PAGANIN']['UNSHARP_COEFF'] == 3.
        self.assertTrue(widget._unsharp_sigma_coeff.text() == str(3.))

        # dilate coefficient
        assert 'DILATE' in self.input['PAGANIN'].all_params
        assert self.input['PAGANIN']['DILATE'] == 2
        self.assertTrue(widget._qleDilatation.text() == str(2))

        # high absorption
        assert 'MKEEP_BONE' in self.input['PAGANIN'].all_params
        assert self.input['PAGANIN']['MKEEP_BONE'] == False
        self.assertFalse(widget._qcbKeepBone.isChecked())

    def testPYHST(self):
        """test that the display of the PyHST widget is correct"""
        widget = self.h5Editor._PyHSTWidget
        widget.load(self.input)

        # checkversion
        assert 'EXE' in self.input['PYHSTEXE'].all_params
        self.assertTrue(widget._qcbPyHSTVersion.currentText() == PyhstRP.OFFV)

        # checkofficialversion
        assert 'OFFV' in self.input['PYHSTEXE'].all_params
        assert self.input['PYHSTEXE']['OFFV'] == PyhstRP.OFFV
        self.assertTrue(widget._qlOfficalVersion.text() == PyhstRP.OFFV)

        # #check verbose
        assert 'VERBOSE' in self.input['PYHSTEXE'].all_params
        assert self.input['PYHSTEXE']['VERBOSE'] == 0
        self.assertFalse(widget._qcbverbose.isChecked())

        # # check output file
        assert 'VERBOSE_FILE' in self.input['PYHSTEXE'].all_params
        assert self.input['PYHSTEXE']['VERBOSE_FILE'] == 'pyhst_out.txt'
        self.assertTrue(widget._qleVerboseFile.text() == 'pyhst_out.txt')

    def testBeamGEO(self):
        """test that the display of the BeamGeo widget is correct"""
        widget = self.h5Editor._beamGeoWidget
        widget.load(self.input)

        # test reconstruction geometry
        assert 'TYPE' in self.input['BEAMGEO'].all_params
        assert self.input['BEAMGEO']['TYPE'] == BeamGeoType.parallel
        self.assertTrue(widget._qcbType.currentText() == 'parallel')

        # test source X
        assert 'SX' in self.input['BEAMGEO'].all_params
        assert self.input['BEAMGEO']['SX'] == 0.
        self.assertTrue(widget._qleSX.text() == str(0.))

        # test source Y
        assert 'SY' in self.input['BEAMGEO'].all_params
        assert self.input['BEAMGEO']['SY'] == 0.
        self.assertTrue(widget._qleSY.text() == str(0.))

        # test source Distance
        assert 'DIST' in self.input['BEAMGEO'].all_params
        assert self.input['BEAMGEO']['DIST'] == 55.
        self.assertTrue(widget._qleDIST.text() == str(55.))

    def testFT(self):
        # parameters from the main widget
        widget = self.h5Editor._mainWidget
        widget.load(self.input)

        assert 'FIXEDSLICE' in self.input['FT'].all_params
        assert self.input['FT']['FIXEDSLICE'].name == 'middle'
        self.assertTrue(widget._qcbSelectSlice.currentText() == 'middle')

        assert 'VOLOUTFILE' in self.input['FT'].all_params
        assert self.input['FT']['VOLOUTFILE'] == 0
        self.assertTrue(widget._qcbEDFStack.isChecked())
        self.assertFalse(widget._qcbSingleVolFile.isChecked())

        assert 'CORRECT_SPIKES_THRESHOLD' in self.input['FT'].all_params
        assert self.input['FT']['CORRECT_SPIKES_THRESHOLD'] == 0.04
        self.assertTrue(widget._qleThresholdSpikesRm.text() == str(0.04))

        assert 'DO_TEST_SLICE' in self.input['FT'].all_params
        assert self.input['FT']['DO_TEST_SLICE'] == 1
        self.assertTrue(widget._qcbDoTestSlice.isChecked())

        assert 'VOLSELECT' in self.input['FT'].all_params
        assert self.input['FT']['VOLSELECT'] == VolSelMode.total
        self.assertTrue(widget._qcbVolumeSelection.currentText() == 'total')

        assert 'VOLSELECTION_REMEMBER' in self.input['FT'].all_params
        assert self.input['FT']['VOLSELECTION_REMEMBER'] == 0
        self.assertFalse(widget._qcbVolSelRemenber.isChecked())

        assert 'RINGSCORRECTION' in self.input['FT'].all_params
        assert self.input['FT']['RINGSCORRECTION'] == 0
        self.assertFalse(widget._qcbRingsCorrection.isChecked())

        # parameters from display widget
        widget = self.h5Editor._displayWidget
        widget.load(self.input)

        assert 'SHOWPROJ' in self.input['FT'].all_params
        assert self.input['FT']['SHOWPROJ'] == 0
        self.assertFalse(widget._qcbShowProj.isChecked())

        assert 'SHOWSLICE' in self.input['FT'].all_params
        assert self.input['FT']['SHOWSLICE'] == 1
        self.assertTrue(widget._qcbShowSlice.isChecked())

        assert 'ANGLE_OFFSET_VALUE' in self.input['FT'].all_params
        assert self.input['FT']['ANGLE_OFFSET_VALUE'] == 0.
        self.assertTrue(widget._qleAngleOffset.text() == str(0.))

        # parameters fron expert widget
        widget = self.h5Editor._expertWidget
        widget.load(self.input)

        assert 'NUM_PART' in self.input['FT'].all_params
        assert self.input['FT']['NUM_PART'] == 4
        self.assertTrue(widget._getNumericalPart() == 4)

        assert 'VERSION' in self.input['FT'].all_params
        assert self.input['FT']['VERSION'] == 'fastomo3 3.2'
        self.assertTrue(widget._qleVersion.text() == 'fastomo3 3.2')

        assert 'DATABASE' in self.input['FT'].all_params
        assert self.input['FT']['DATABASE'] == 0
        self.assertFalse(widget._qcbDataBase.isChecked())

        assert 'NO_CHECK' in self.input['FT'].all_params
        assert self.input['FT']['NO_CHECK'] == 0
        self.assertFalse(widget._qcbNocheck.isChecked())

        assert 'ZEROOFFMASK' in self.input['FT'].all_params
        assert self.input['FT']['ZEROOFFMASK'] == 1
        self.assertTrue(widget._qcbZeroRegionMask.isChecked())

        assert 'FIXHD' in self.input['FT'].all_params
        assert self.input['FT']['FIXHD'] == 0
        self.assertFalse(widget._qcbFixHD.isChecked())

    def testFTAxis(self):
        widget = self.h5Editor._axisWidget
        widget.load(self.input)

        self.assertTrue('DO_AXIS_CORRECTION' in self.input['FTAXIS'].all_params)
        self.assertEqual(self.input['FTAXIS']['DO_AXIS_CORRECTION'], True)
        widget._doAxisCorrection.setChecked(False)
        self.assertTrue(self.input['FTAXIS']['DO_AXIS_CORRECTION'], False)

        self.assertTrue('USE_TOMWER_AXIS' in self.input['FTAXIS'].all_params)
        self.assertEqual(self.input['FTAXIS']['USE_TOMWER_AXIS'], True)
        widget._useTomwerAxis.setChecked(False)
        self.assertTrue(self.input['FTAXIS']['USE_TOMWER_AXIS'], False)

        self.assertTrue('TRY_USE_OLD_TOMWER_AXIS' in self.input['FTAXIS'].all_params)
        self.assertEqual(self.input['FTAXIS']['TRY_USE_OLD_TOMWER_AXIS'], True)
        widget._useOldTomwerAxis.setChecked(False)
        self.assertTrue(self.input['FTAXIS']['TRY_USE_OLD_TOMWER_AXIS'], False)

        self.assertTrue('HALF_ACQ' in self.input['FT'].all_params)
        self.assertEqual(self.input['FT']['HALF_ACQ'], False)
        widget._qcbHalfAcq.isChecked()
        self.assertTrue(self.input['FTAXIS']['TRY_USE_OLD_TOMWER_AXIS'], True)

    def testDKRF(self):
        widget = self.h5Editor._dkRefWidget
        widget.load(self.input)

        assert 'DARKCAL' in self.input['DKRF'].all_params
        assert self.input['DKRF']['DARKCAL'] == dkrf_method.average
        self.assertTrue(widget._qcbDKMode.getMode() == dkrf_method.average)

        assert 'REFSCAL' in self.input['DKRF'].all_params
        assert self.input['DKRF']['REFSCAL'] == dkrf_method.median
        self.assertTrue(widget._qcbRefMode.getMode() == dkrf_method.median)

        assert 'DKFILE' in self.input['DKRF'].all_params
        assert self.input['DKRF']['DKFILE'] == 'darkend[0-9]{3,4}'
        self.assertTrue(widget._qleDKPattern.text() == 'darkend[0-9]{3,4}')

        assert 'RFFILE' in self.input['DKRF'].all_params
        assert self.input['DKRF']['RFFILE'] == 'ref*.*[0-9]{3,4}_[0-9]{3,4}'
        self.assertTrue(widget._qleRefsPattern.text() == 'ref*.*[0-9]{3,4}_[0-9]{3,4}')

        assert 'REFSRMV' in self.input['DKRF'].all_params
        assert self.input['DKRF']['REFSRMV'] == 0
        self.assertFalse(widget._qcbRmRef.isChecked())

        assert 'REFSOVE' in self.input['DKRF'].all_params
        assert self.input['DKRF']['REFSOVE'] == 0
        self.assertTrue(widget._qcbSkipRef.isChecked())


@unittest.skipIf(skip_gui_test(), reason='skip gui test')
class TestSimplifyH5EditorSave(TestCaseQt):
    """test that h5editor is returning the correct H5 structures"""

    def setUp(self):
        super(TestSimplifyH5EditorSave, self).setUp()
        self.h5Editor = ReconsParamsEditor(None)
        self.h5Editor.setAttribute(qt.Qt.WA_DeleteOnClose)

        self.input = QReconsParams()
        self.h5Editor.loadReconsParams(self.input)

    def tearDown(self):
        self.qapp.processEvents()
        self.h5Editor.close()
        del self.h5Editor
        self.qapp.processEvents()
        super(TestSimplifyH5EditorSave, self).tearDown()

    def testSaving(self):
        # changing some values in BEAMGEO
        assert self.input.to_dict()['BEAMGEO']['TYPE'] == 'p'
        self.input['BEAMGEO']['TYPE'] = 'c'
        # changing some values in PAGANIN
        assert self.input['PAGANIN']['MKEEP_BONE'] == 0
        self.input['PAGANIN']['MKEEP_BONE'] = 1
        # changing some values in PYHSTEXE
        assert self.input['PYHSTEXE']['VERBOSE_FILE'] == 'pyhst_out.txt'
        self.input['PYHSTEXE']['VERBOSE_FILE'] = 'pyhst_out2.txt'
        assert self.input['PYHSTEXE']['OFFV'] == PyhstRP.OFFV
        self.input['PYHSTEXE']['OFFV'] = 'randomString'
        # changing some FT value
        self.input['FT']['CORRECT_SPIKES_THRESHOLD'] = 'Inf'
        self.input['FT']['RINGSCORRECTION'] = 1
        self.input['FT']['ANGLE_OFFSET_VALUE'] = 25.0
        self.input['FT']['FIXEDSLICE'] = '12'
        self.input['FT']['SHOWSLICE'] = 0

        # changing Paganin values
        self.input['PAGANIN']['MODE'] = 3

        # changing the DKRF values
        self.input['DKRF']['REFSCAL'] = 'Average'
        self.input['DKRF']['DARKCAL'] = 'None'
        self.input['DKRF']['RFFILE'] = 'toto.*'
        self.input['DKRF']['REFSRMV'] = 0

        # loading from new values
        self.h5Editor.loadReconsParams(self.input)

        # changing again some values
        self.h5Editor._PaganinWidget._qcbpaganin.setCurrentIndex(1)
        self.h5Editor._PyHSTWidget._qcbverbose.setCheckState(qt.Qt.Unchecked)
        self.h5Editor._beamGeoWidget._qleSY.setText('12.0')
        # add some checking on the FT
        self.assertFalse(self.h5Editor._mainWidget._grpThreshold.isChecked())  # this shouldn't be activated

        # saving the values
        savedStructures = self.h5Editor.getStructs()
        self.assertTrue('BEAMGEO' in savedStructures)
        self.assertTrue('PYHSTEXE' in savedStructures)
        self.assertTrue('PAGANIN' in savedStructures)
        self.assertTrue('FT' in savedStructures)
        self.assertTrue('FTAXIS' in savedStructures)
        self.assertTrue(savedStructures['BEAMGEO']['TYPE'] == 'c')
        self.assertTrue(savedStructures['BEAMGEO']['SY'] == 12.0)

        self.assertTrue(savedStructures['PAGANIN']['MKEEP_BONE'] == 1)
        self.assertTrue(savedStructures['PAGANIN']['MODE'] == 1)

        self.assertTrue(savedStructures['PYHSTEXE']['VERBOSE_FILE'] == 'pyhst_out2.txt')
        self.assertTrue(savedStructures['PYHSTEXE']['OFFV'] == 'randomString')
        self.assertTrue(savedStructures['PYHSTEXE']['VERBOSE'] == 0)

        self.assertTrue(savedStructures['FT']['CORRECT_SPIKES_THRESHOLD'] == 'Inf')
        self.assertTrue(savedStructures['FT']['RINGSCORRECTION'] == 1)
        self.assertTrue(savedStructures['FT']['ANGLE_OFFSET'] == 1)
        self.assertEqual(savedStructures['FT']['FIXEDSLICE'], '12')
        self.assertTrue(savedStructures['FT']['SHOWSLICE'] == 0)

        self.assertTrue(savedStructures['FTAXIS']['DO_AXIS_CORRECTION'] == 1)


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestSimplifyH5EditorDisplay, TestSimplifyH5EditorSave, ):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite


if __name__ == '__main__':
    unittest.main(defaultTest="suite")



# coding: utf-8
#/*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

"""
This module is containing benchmarks for rsync
"""

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "07/06/2017"


import os
import shutil
import tempfile
import time

from tomwer.synctools.rsyncmanager import RSyncManager


class BenchRSyncClass(object):
    ORI_FOLDER = '/home/payno/testRSync/data/'

    def __init__(self):
        """
        """
        self.folder1Path = tempfile.mkdtemp()
        print(self.folder1Path)
        self.folder2Path = tempfile.mkdtemp()

        assert (os.path.isdir(self.folder1Path))
        assert (os.path.isdir(self.folder2Path))
        # assert (os.path.isdir(BenchRSyncClass.ORI_FOLDER))
        # copy from original to folder 1
        # shutil.copytree(src=BenchRSyncClass.ORI_FOLDER,
        #                dst=os.path.join(self.folder1Path, 'test'))

    def __del__(self):
        if os.path.isdir(self.folder1Path):
            shutil.rmtree(self.folder1Path)

        if os.path.isdir(self.folder2Path):
            shutil.rmtree(self.folder2Path)

    def testCopy(self, options, parallel=False, nbIter=3):

        startTime = time.time()

        for it in range(nbIter):
            # copy from folder 1 to folder 2
            RSyncManager().syncFolderRaw(source=self.folder1Path,
                                         target=self.folder2Path,
                                         options=options,
                                         parallel=parallel,
                                         block=True)

            # copy from folder 2 to folder 1
            RSyncManager().syncFolderRaw(source=self.folder2Path,
                                         target=self.folder1Path,
                                         options=options,
                                         parallel=parallel,
                                         block=True)

        # check data integrity ?
        return (time.time() - startTime) / nbIter

    def testCopyFolders(self, f1, f2, options, parallel=False, nbIter=1):

        fullTime = 0

        for it in range(nbIter):
            startTime = time.time()

            # copy from folder 1 to folder 2
            RSyncManager().syncFolderRaw(source=f1,
                                         target=f2,
                                         options=options,
                                         parallel=parallel,
                                         block=True)

            fullTime += (time.time() - startTime)

            # copy from folder 2 to folder 1
            RSyncManager().syncFolderRaw(source=f2,
                                         target=f1,
                                         options=options,
                                         parallel=parallel,
                                         block=True)

        return (fullTime / nbIter)

    def testCopyTree(self, parallel=True):

        startTime = time.time()

        # copy from original to folder 1
        shutil.copytree(src=self.folder1Path,
                        dst=os.path.join(self.folder2Path, 'test'))

        shutil.rmtree(os.path.join(self.folder1Path, 'test'))
        # copy from original to folder 1
        shutil.copytree(src=self.folder2Path,
                        dst=os.path.join(self.folder1Path, 'test'))

        # check data integrity ?
        return (time.time() - startTime)


def benchemark_rsync():
    l = BenchRSyncClass()

    options = []
    options.append(['-a'])
    options.append(['-r'])

    f1 = '/lbsram/data/visitor/mi1226/bm05/nemoz/test1_'
    f2 = '/data/visitor/mi1226/bm05/nemoz/test1_'

    print('----- not using parallel ------')

    for opt in options:
        print('options are : ')
        print(opt)
        print(l.testCopyFolders(f1, f2, opt, parallel=False))

#    print('----- using parallel ------')

#    for opt in options:
#        print('options are : ')
#        print(opt)
#        print(l.testCopyFolders(f1, f2, opt, parallel=True))

    print('----- default shutil ------')
    print(l.testCopyTree())


if __name__ == '__main__':
    benchemark_rsync()



from silx.gui import qt
from Orange.canvas import config, __main__ as main
from .splash import splash_screen, getIcon
import tomwer.version
import os, sys
from Orange.misc import environ
import logging


def version():
    return tomwer.version.version


class TomwerConfig(config.Config):
    ApplicationName = "tomwer"
    ApplicationVersion = version()

    @staticmethod
    def splash_screen():
        return splash_screen()

    @staticmethod
    def core_packages():
        return super(TomwerConfig, TomwerConfig).core_packages() + [
            "tomwer-add-on",
        ]

    @staticmethod
    def application_icon():
        return getIcon()


class TomwerSplashScreen(qt.QSplashScreen):
    """SplashScreen to overwrite the one of Orange"""
    def __init__(self, parent=None, pixmap=None, textRect=None,
                 textFormat=qt.Qt.PlainText, **kwargs):
        qt.QSplashScreen.__init__(self, parent, pixmap=pixmap, **kwargs)

    def showMessage(self, message, alignment=qt.Qt.AlignLeft, color=qt.Qt.black):
        version = 'version ' + str(tomwer.version.version)
        super().showMessage(version, qt.Qt.AlignLeft | qt.Qt.AlignBottom, qt.Qt.black)


class Launcher:
    """Proxy to orange-canvas"""
    def launch(self, argv):
        config.Config = TomwerConfig
        self.fix_application_dirs()
        self.replace_splash_screen()
        self.main(argv)

    def fix_application_dirs(self):
        def data_dir(versioned=True):
            """
            Return the platform dependent Orange data directory.

            This is ``data_dir_base()``/Orange/__VERSION__/ directory if versioned is
            `True` and ``data_dir_base()``/Orange/ otherwise.
            """
            base = environ.data_dir_base()
            if versioned:
                return os.path.join(base, "tomwer", version())
            else:
                return os.path.join(base, "tomwer")

        environ.data_dir = data_dir

        def cache_dir(*args):
            """
            Return the platform dependent Orange cache directory.
            """
            if sys.platform == "darwin":
                base = os.path.expanduser("~/Library/Caches")
            elif sys.platform == "win32":
                base = os.getenv("APPDATA", os.path.expanduser("~/AppData/Local"))
            elif os.name == "posix":
                base = os.getenv("XDG_CACHE_HOME", os.path.expanduser("~/.cache"))
            else:
                base = os.path.expanduser("~/.cache")

            base = os.path.join(base, "tomwer", version())
            if sys.platform == "win32":
                # On Windows cache and data dir are the same.
                # Microsoft suggest using a Cache subdirectory
                return os.path.join(base, "Cache")
            else:
                return base

        environ.cache_dir = cache_dir

    def replace_splash_screen(self):
        main.SplashScreen = TomwerSplashScreen

    def main(self, argv):
        from Orange.canvas.__main__ import main
        logging.getLogger('tomoscan.esrf.hdf5scan').setLevel(logging.WARNING)
        logging.getLogger('tomoscan.esrf.edfscan').setLevel(logging.WARNING)

        main(argv)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "24/01/2017"


from tomwer.synctools.datatransfert import FolderTransfert
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.utils import rebaseParFile
from tomwer.core.utils.scanutils import MockEDF
from tomwer.test.utils import UtilsTest
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from nxtomomill import converter as nxtomomill_converter
from silx.gui.utils.testutils import TestCaseQt
from glob import glob
import unittest
import os
import tempfile
import shutil
import logging
try:
    from tomwer.synctools.rsyncmanager import RSyncManager
except ImportError:
    has_rsync = False
else:
    has_rsync = True

logging.disable(logging.INFO)


class TestEDFDataTransfert(TestCaseQt):
    """
    Test that the folder transfert process is valid
    """

    def setUp(self):
        TestCaseQt.setUp(self)
        self.sourcedir = tempfile.mkdtemp()
        self.n_file = 10
        MockEDF.fastMockAcquisition(self.sourcedir, n_radio=self.n_file)
        self.scan = ScanFactory.create_scan_object(self.sourcedir)
        assert(os.path.isdir(self.sourcedir))
        self.targettedir = tempfile.mkdtemp()
        assert(os.path.isdir(self.targettedir))

        self.folderTrans = FolderTransfert()
        self.folderTrans.turn_off_print = True
        self.folderTrans.setDestDir(self.targettedir)
        self.folderTrans._copying = False

        self.outputdir = os.path.join(self.targettedir, os.path.basename(self.sourcedir))

    def tearDown(self):
        if os.path.isdir(self.sourcedir):
            shutil.rmtree(self.sourcedir)
        if os.path.isdir(self.targettedir):
            shutil.rmtree(self.targettedir)
        TestCaseQt.tearDown(self)

    def testMoveFiles(self):
        """
        simple test that files are moved
        """
        self.folderTrans.process(self.scan,
                                 move=True,
                                 noRsync=True)


        self.assertTrue(os.path.isdir(self.outputdir))
        self.assertTrue(self.checkDataCopied())

    def testCopyFiles(self):
        """
        Simple test that file are copy and deleted
        """
        self.folderTrans.process(self.scan,
                                 move=False,
                                 noRsync=True)

        self.assertTrue(self.checkDataCopied())

    def testMoveFilesForce(self):
        """
        Test the force option of folderTransfert
        """
        assert(not os.path.isdir(self.outputdir))
        assert(os.path.isdir(self.scan.path))
        self.folderTrans.process(self.scan,
                                 move=True,
                                 force=False,
                                 noRsync=True)

        MockEDF.fastMockAcquisition(self.sourcedir, n_radio=self.n_file)
        with self.assertRaises(shutil.Error):
            self.assertRaises(
                self.folderTrans.process(self.scan,
                                         move=True,
                                         force=False,
                                         noRsync=True))

        self.folderTrans.process(self.scan,
                                 move=True,
                                 force=True,
                                 noRsync=True)
        self.assertTrue(self.checkDataCopied())

    def testCopyFilesForce(self):
        """
        Test the force option for the copy files process
        """
        assert(not os.path.isdir(self.outputdir))
        os.mkdir(self.outputdir)
        assert(os.path.isdir(self.outputdir))
        self.folderTrans.process(self.scan,
                                 move=False,
                                 force=False,
                                 noRsync=True)
        self.assertTrue(self.checkDataCopied())

        MockEDF.fastMockAcquisition(self.sourcedir, n_radio=self.n_file)
        self.assertTrue(self.scan.path == self.sourcedir)
        with self.assertRaises(FileExistsError):
            self.assertRaises(
                self.folderTrans.process(self.scan,
                                         move=False,
                                         force=False,
                                         noRsync=True))

        self.folderTrans.process(self.scan,
                                 move=False,
                                 force=True,
                                 noRsync=True)
        self.assertTrue(self.checkDataCopied())

    def checkDataCopied(self):
        outputFiles = os.listdir(self.outputdir)
        inputFile = glob(self.sourcedir)
        # + 3 because .info and .xml are count + one tomwer_processes file
        return (len(inputFile) == 0) and (len(outputFiles) == (self.n_file + 3)) and \
            (not os.path.isdir(self.sourcedir))


@unittest.skipIf(RSyncManager().has_rsync() is False, "Rsync is missing")
class TestHDFDataTransfert(unittest.TestCase):
    """Make sure we can transfer data from bliss acquisition"""

    def setUp(self):
        self.input_dir = tempfile.mkdtemp()
        self.output_dir = tempfile.mkdtemp()
        shutil.copytree(UtilsTest.getBlissDataset(folderID='sample'),
                        os.path.join(self.input_dir, 'sample'))

        self._proposal_file = os.path.join(self.input_dir, 'sample',
                                           'ihpayno_sample.h5')
        assert os.path.exists(self._proposal_file)
        self._sample_file = os.path.join(self.input_dir, 'sample',
                                         'sample_29042021',
                                         'sample_29042021.h5')
        assert os.path.exists(self._sample_file)
        self._sample_file_entry = '/1.1'

        self.process = FolderTransfert()
        self.process.setDestDir(os.path.join(self.output_dir, 'sample'))

        output_file_path = os.path.join(self.input_dir, 'sample',
                                        'sample_29042021',
                                        'nx_file.nx')
        # convert it to nx
        files_entries = nxtomomill_converter.h5_to_nx(input_file_path=self._sample_file,
                                                      output_file=output_file_path,
                                                      entries=(self._sample_file_entry,),
                                                      single_file=True,
                                                      ask_before_overwrite=False,
                                                      request_input=False,
                                                      file_extension='.nx')
        assert len(files_entries) == 1
        output_file_path, entry = files_entries[0]
        assert os.path.exists(output_file_path)
        assert os.path.isfile(output_file_path)
        self.scan = HDF5TomoScan(scan=output_file_path,
                                 entry=entry)
        self.process.setForceSync(True)

        # register information regarding origin of the file
        from tomwer.core.process.datalistener import DataListener
        source_scans = DataListener._get_scan_sources(self._sample_file,
                                                      self._sample_file_entry)
        DataListener._register_process(process_file=self.scan.process_file,
                                       process=DataListener,
                                       entry=self.scan.entry,
                                       results={'output_file': self.scan.master_file,
                                                'entry': self.scan.entry},
                                       configuration={'sample_file': self._sample_file,
                                                      'entry': self._sample_file_entry,
                                                      'source_scans': source_scans,
                                                      'file_proposal': self._proposal_file,
                                                      },
                                       process_index=self.scan.pop_process_index(),
                                       overwrite=True)

    def tearDown(self):
        self.process.setForceSync(False)
        shutil.rmtree(self.input_dir)
        shutil.rmtree(self.output_dir)

    def testDataTransfert(self):
        self.process.process(scan=self.scan)
        output_scan = os.path.join(self.output_dir, 'sample',
                                   'sample_29042021', 'scan0002')
        output_proposal_file = os.path.join(self.output_dir, 'sample',
                                            'ihpayno_sample.h5')
        output_sample_file = os.path.join(self.output_dir, 'sample',
                                          'sample_29042021',
                                          'sample_29042021.h5')
        self.assertTrue(os.path.exists(output_scan))
        self.assertEqual(len(os.listdir(output_scan)), 3)
        self.assertTrue(os.path.exists(output_proposal_file))
        self.assertTrue(os.path.exists(output_sample_file))
        other_scan_dir = os.path.join(self.output_dir, 'sample', 'scan0004')
        self.assertFalse(os.path.exists(other_scan_dir))


class TestPreTransfert(unittest.TestCase):
    """Test the pretransfert actions"""
    def setUp(self):
        unittest.TestCase.setUp(self)
        folderDataset = UtilsTest.getEDFDataset('scan_3_')
        self.tmpdir = tempfile.mkdtemp()
        self.outputfolder = tempfile.mkdtemp()
        scan_path = os.path.join(self.tmpdir, 'scan_3_')
        shutil.copytree(src=folderDataset, dst=scan_path)
        self.scan = ScanFactory.create_scan_object(scan_path=scan_path)
        self.folderTrans = FolderTransfert()
        self.folderTrans.turn_off_print = True
        self.folderTrans._copying = False
        self.folderTrans.setDestDir(self.outputfolder)

    def tearDown(self):
        unittest.TestCase.tearDown(self)
        for _dir in (self.tmpdir, self.outputfolder):
            if os.path.exists(_dir):
                shutil.rmtree(_dir)

    def testParFile(self):
        """Make sure the scan_3_.par is correctly updated when moved
        """
        parDict = self.getParDict(os.path.join(self.scan.path, 'scan_3_.par'))
        assert parDict['file_prefix'] == '/data/visitor/mi1226/id19/nemoz/henri/scan_3_/scan_3_'
        assert parDict['ff_prefix'] == '/data/visitor/mi1226/id19/nemoz/henri/scan_3_/refHST'
        rebaseParFile(os.path.join(self.scan.path, 'scan_3_.par'),
                      oldfolder='/data/visitor/mi1226/id19/nemoz/henri/scan_3_',
                      newfolder=self.scan.path)
        parDict = self.getParDict(os.path.join(self.scan.path, 'scan_3_.par'))
        self.assertTrue(parDict['file_prefix'] == os.path.join(self.scan.path, 'scan_3_'))
        self.assertTrue(parDict['ff_prefix'] == os.path.join(self.scan.path, 'refHST'))
        self.folderTrans.process(scan=self.scan, move=False, force=True,
                                 noRsync=True)
        parDict = self.getParDict(os.path.join(self.outputfolder, 'scan_3_', 'scan_3_.par'))
        self.assertTrue(parDict['file_prefix'] == os.path.join(self.outputfolder, 'scan_3_', 'scan_3_'))
        self.assertTrue(parDict['ff_prefix'] == os.path.join(self.outputfolder, 'scan_3_', 'refHST'))

    @staticmethod
    def getParDict(_file):
        assert os.path.isfile(_file)
        ddict = {}
        f = open(_file, "r")
        lines = f.readlines()
        for line in lines:
            if not '=' in line:
                continue
            l = line.rstrip().replace(' ', '')
            l = l.split('#')[0]
            key, value = l.split('=')
            ddict[key.lower()] = value
        return ddict


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestEDFDataTransfert, TestHDFDataTransfert, TestPreTransfert):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite


if __name__ == '__main__':
    unittest.main(defaultTest="suite")
